import logging
import sys
from datetime import timedelta
from typing import Dict, Optional

from PyQt6 import QtCore, uic
from PyQt6.QtCore import QItemSelectionModel, QMimeData, QPoint, Qt, pyqtSlot
from PyQt6.QtGui import QDesktopServices, QKeySequence, QShortcut
from PyQt6.QtWidgets import (
    QAbstractItemView,
    QApplication,
    QHeaderView,
    QLayout,
    QMenu,
    QMessageBox,
    QStyledItemDelegate,
    QTableView,
    QTableWidgetItem,
    QWidget,
)

from vorta.borg.check import BorgCheckJob
from vorta.borg.compact import BorgCompactJob
from vorta.borg.delete import BorgDeleteJob
from vorta.borg.diff import BorgDiffJob
from vorta.borg.info_archive import BorgInfoArchiveJob
from vorta.borg.list_repo import BorgListRepoJob
from vorta.borg.prune import BorgPruneJob
from vorta.borg.rename import BorgRenameJob
from vorta.i18n import trans_late, translate
from vorta.i18n.richtext import escape, format_richtext, link
from vorta.store.models import ArchiveModel, SettingsModel
from vorta.utils import (
    borg_compat,
    find_best_unit_for_sizes,
    format_archive_name,
    get_asset,
    get_mount_points,
    pretty_bytes,
)
from vorta.views.archive.archive_extract import ArchiveExtract
from vorta.views.archive.archive_mount import ArchiveMount
from vorta.views.base_tab import BaseTab
from vorta.views.dialogs.archive import diff_result
from vorta.views.dialogs.archive.diff_result import DiffResultDialog, DiffTree
from vorta.views.source_tab import SizeItem
from vorta.views.utils import get_colored_icon

uifile = get_asset('UI/archive_tab.ui')
ArchiveTabUI, ArchiveTabBase = uic.loadUiType(uifile)

logger = logging.getLogger(__name__)


#: The number of decimal digits to show in the size column
SIZE_DECIMAL_DIGITS = 1


# from https://stackoverflow.com/questions/63177587/pyqt-tableview-align-icons-to-center
class IconDelegate(QStyledItemDelegate):
    def initStyleOption(self, option, index):
        super().initStyleOption(option, index)
        option.decorationSize = option.rect.size() - QtCore.QSize(0, 10)


class ArchiveTab(BaseTab, ArchiveTabBase, ArchiveTabUI):
    prune_intervals = ['hour', 'day', 'week', 'month', 'year']

    def __init__(self, parent=None, app=None, profile_provider=None):
        """Init."""
        super().__init__(parent=parent, profile_provider=profile_provider)
        self.setupUi(parent)
        self.is_editing = False  # track if the cell edit was completed or canceled
        self.mount_points = {}  # mapping of archive name to mount point
        self.repo_mount_point: Optional[str] = None  # mount point of whole repo
        self.menu = None
        self.app = app or self.app
        self.toolBox.setCurrentIndex(0)
        self.repoactions_enabled = True
        self.renamed_archive_original_name = None
        self.remaining_refresh_archives = (
            0  # number of archives that are left to refresh before action buttons are enabled again
        )

        self.archive_mount = ArchiveMount(self)
        self.archive_extract = ArchiveExtract(self)

        #: Tooltip dict to save the tooltips set in the designer
        self.tooltip_dict: Dict[QWidget, str] = {}
        self.tooltip_dict[self.bDiff] = self.bDiff.toolTip()
        self.tooltip_dict[self.bDelete] = self.bDelete.toolTip()
        self.tooltip_dict[self.bRefreshArchive] = self.bRefreshArchive.toolTip()
        self.tooltip_dict[self.compactButton] = self.compactButton.toolTip()

        header = self.archiveTable.horizontalHeader()
        header.setVisible(True)
        header.setSectionResizeMode(0, QHeaderView.ResizeMode.ResizeToContents)
        header.setSectionResizeMode(1, QHeaderView.ResizeMode.ResizeToContents)
        header.setSectionResizeMode(2, QHeaderView.ResizeMode.ResizeToContents)
        header.setSectionResizeMode(3, QHeaderView.ResizeMode.Interactive)
        header.setSectionResizeMode(4, QHeaderView.ResizeMode.Stretch)
        header.setSectionResizeMode(5, QHeaderView.ResizeMode.ResizeToContents)

        delegate = IconDelegate(self.archiveTable)
        self.archiveTable.setItemDelegateForColumn(5, delegate)

        self.mount_help_text = trans_late('Form', 'To mount archives, first install "FUSE for macOS" from %1.')
        self.mount_help_link_text = trans_late('Form', 'here')
        self.pruning_help_text = trans_late(
            'Form',
            'Pruning removes older archives. You can choose the number of hourly, daily, etc. archives to preserve. '
            'Usually you will keep more newer and fewer old archives. Read %1.',
        )
        self.pruning_help_link_text = trans_late('Form', 'more')
        self._init_help_texts()

        if sys.platform != 'darwin':
            self._set_status('')  # Set platform-specific hints.

        self.archiveTable.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.archiveTable.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
        self.archiveTable.setWordWrap(False)
        self.archiveTable.setTextElideMode(QtCore.Qt.TextElideMode.ElideLeft)
        self.archiveTable.setAlternatingRowColors(True)
        self.archiveTable.cellDoubleClicked.connect(self.cell_double_clicked)
        self.archiveTable.setSortingEnabled(True)
        self.archiveTable.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.archiveTable.customContextMenuRequested.connect(self.archiveitem_contextmenu)
        edit_delegate = self.archiveTable.itemDelegate()
        edit_delegate.closeEditor.connect(self.on_editing_finished)

        # shortcuts
        shortcut_copy = QShortcut(QKeySequence.StandardKey.Copy, self.archiveTable)
        shortcut_copy.activated.connect(self.archive_copy)

        # single and double selection feature
        self.archiveTable.setSelectionMode(QTableView.SelectionMode.ExtendedSelection)
        self.archiveTable.selectionModel().selectionChanged.connect(self.on_selection_change)

        # connect archive actions
        self.bMountArchive.clicked.connect(self.archive_mount.bmountarchive_clicked)
        self.bRefreshArchive.clicked.connect(self.refresh_archive_info)
        self.bRename.clicked.connect(self.cell_double_clicked)
        self.bDelete.clicked.connect(self.delete_action)
        self.bExtract.clicked.connect(self.archive_extract.extract_action)
        self.compactButton.clicked.connect(self.compact_action)

        # other signals
        self.bList.clicked.connect(self.refresh_archive_list)
        self.bPrune.clicked.connect(self.prune_action)
        self.bCheck.clicked.connect(self.check_action)
        self.bDiff.clicked.connect(self.diff_action)
        self.bMountRepo.clicked.connect(self.archive_mount.bmountrepo_clicked)

        self.archiveNameTemplate.textChanged.connect(
            lambda tpl, key='new_archive_name': self.save_archive_template(tpl, key)
        )
        self.prunePrefixTemplate.textChanged.connect(
            lambda tpl, key='prune_prefix': self.save_archive_template(tpl, key)
        )

        # Connect prune setting signals (only once in init, not in populate_from_profile)
        for i in self.prune_intervals:
            getattr(self, f'prune_{i}').valueChanged.connect(self.save_prune_setting)
        self.prune_keep_within.editingFinished.connect(self.save_prune_setting)

        self.track_profile_change(call_now=True)
        self.set_icons()

        # Connect to events
        self.track_palette_change()
        self.track_palette_change(self.populate_from_profile)
        self.track_backup_finished()
        self.track_profile_change(self.toggle_compact_button_visibility)
        self.track_signal(self.app.backup_cancelled_event, self.cancel_action)

    def set_icons(self):
        """Used when changing between light- and dark mode"""
        self.bCheck.setIcon(get_colored_icon('check-circle'))
        self.bDiff.setIcon(get_colored_icon('stream-solid'))
        self.bPrune.setIcon(get_colored_icon('cut'))
        self.bList.setIcon(get_colored_icon('refresh'))
        self.compactButton.setIcon(get_colored_icon('broom-solid'))
        self.toolBox.setItemIcon(0, get_colored_icon('tasks'))
        self.toolBox.setItemIcon(1, get_colored_icon('cut'))
        self.bRefreshArchive.setIcon(get_colored_icon('refresh'))
        self.bRename.setIcon(get_colored_icon('edit'))
        self.bDelete.setIcon(get_colored_icon('trash'))
        self.bExtract.setIcon(get_colored_icon('cloud-download'))

        self.archive_mount.bmountarchive_refresh(icon_only=True)
        self.archive_mount.bmountrepo_refresh()

    @pyqtSlot(QPoint)
    def archiveitem_contextmenu(self, pos: QPoint):
        # index under cursor
        index = self.archiveTable.indexAt(pos)
        if not index.isValid():
            return  # popup only for items

        selected_rows = self.archiveTable.selectionModel().selectedRows(index.column())

        if selected_rows and index not in selected_rows:
            return  # popup only for selected items

        menu = QMenu(self.archiveTable)
        menu.addAction(get_colored_icon('copy'), self.tr("Copy"), lambda: self.archive_copy(index=index))
        menu.addSeparator()

        # archive actions
        button_connection_pairs = [
            (self.bRefreshArchive, self.refresh_archive_info),
            (self.bDiff, self.diff_action),
            (self.bMountArchive, self.archive_mount.bmountarchive_clicked),
            (self.bExtract, self.archive_extract.extract_action),
            (self.bRename, self.cell_double_clicked),
            (self.bDelete, self.delete_action),
        ]

        for button, connection in button_connection_pairs:
            action = menu.addAction(button.icon(), button.text(), connection)
            action.setEnabled(button.isEnabled())

        menu.popup(self.archiveTable.viewport().mapToGlobal(pos))

    def cancel_action(self):
        self._set_status(self.tr("Action cancelled."))
        self._toggle_all_buttons(True)

    def _set_status(self, text):
        self.mountErrors.setText(text)
        self.mountErrors.repaint()

    def _init_help_texts(self):
        mount_template = self.mountErrors.text()
        mount_sentence = format_richtext(
            escape(translate('Form', self.mount_help_text)),
            link('https://macfuse.github.io/', translate('Form', self.mount_help_link_text)),
        )
        self.mountErrors.setText(format_richtext(mount_template, mount_sentence))

        prune_template = self.pruningHelpLabel.text()
        prune_sentence = format_richtext(
            escape(translate('Form', self.pruning_help_text)),
            link(
                'https://borgbackup.readthedocs.io/en/stable/usage/prune.html',
                translate('Form', self.pruning_help_link_text),
                color="#FF4500",
            ),
        )
        self.pruningHelpLabel.setText(format_richtext(prune_template, prune_sentence))

    def _toggle_all_buttons(self, enabled=True):
        """
        Set all the buttons in the archive panel to the given state.

        Parameters
        ----------
        enabled : bool, optional
            The enabled state, by default True
        """
        self.repoactions_enabled = enabled

        for button in [
            self.bCheck,
            self.bList,
            self.bPrune,
            self.bDiff,
            self.bMountRepo,
            self.bDelete,
            self.compactButton,
            self.fArchiveActions,
        ]:
            button.setEnabled(enabled)
            button.repaint()

        # Restore states
        self.on_selection_change()

    def populate_from_profile(self):
        """Populate archive list and prune settings from profile."""
        profile = self.profile()
        if profile.repo is not None:
            # get mount points
            self.mount_points, repo_mount_points = get_mount_points(profile.repo.url)
            if repo_mount_points:
                self.repo_mount_point = repo_mount_points[0]

            if profile.repo.name:
                repo_name = f"{profile.repo.name} ({profile.repo.url})"
            else:
                repo_name = profile.repo.url
            self.toolBox.setItemText(0, self.tr('Archives for {}').format(repo_name))

            archives = [s for s in profile.repo.archives.select().order_by(ArchiveModel.time.desc())]

            # if no archive's name can be found in self.mount_points, then hide the mount point column
            if not any(a.name in self.mount_points for a in archives):
                self.archiveTable.hideColumn(3)
            else:
                self.archiveTable.showColumn(3)

            sorting = self.archiveTable.isSortingEnabled()
            self.archiveTable.setSortingEnabled(False)
            best_unit = find_best_unit_for_sizes((a.size for a in archives), precision=SIZE_DECIMAL_DIGITS)
            for row, archive in enumerate(archives):
                self.archiveTable.insertRow(row)

                formatted_time = archive.time.strftime('%Y-%m-%d %H:%M')
                self.archiveTable.setItem(row, 0, QTableWidgetItem(formatted_time))

                # format units based on user settings for 'dynamic' or 'fixed' units
                fixed_unit = best_unit if SettingsModel.get(key='enable_fixed_units').value else None
                size = pretty_bytes(archive.size, fixed_unit=fixed_unit, precision=SIZE_DECIMAL_DIGITS)
                self.archiveTable.setItem(row, 1, SizeItem(size))

                if archive.duration is not None:
                    formatted_duration = str(timedelta(seconds=round(archive.duration)))
                else:
                    formatted_duration = ''

                self.archiveTable.setItem(row, 2, QTableWidgetItem(formatted_duration))

                mount_point = self.mount_points.get(archive.name)
                if mount_point is not None:
                    item = QTableWidgetItem(mount_point)
                    self.archiveTable.setItem(row, 3, item)

                self.archiveTable.setItem(row, 4, QTableWidgetItem(archive.name))

                if archive.trigger == 'scheduled':
                    item = QTableWidgetItem(get_colored_icon('clock-o'), '')
                    item.setToolTip(self.tr('Scheduled'))
                    self.archiveTable.setItem(row, 5, item)
                elif archive.trigger == 'user':
                    item = QTableWidgetItem(get_colored_icon('user'), '')
                    item.setToolTip(self.tr('User initiated'))
                    item.setTextAlignment(Qt.AlignmentFlag.AlignRight)
                    self.archiveTable.setItem(row, 5, item)

            self.archiveTable.setRowCount(len(archives))
            self.archiveTable.setSortingEnabled(sorting)
            item = self.archiveTable.item(0, 0)
            self.archiveTable.scrollToItem(item)

            self.archiveTable.selectionModel().clearSelection()
            if self.remaining_refresh_archives == 0:
                self._toggle_all_buttons(enabled=True)
        else:
            self.mount_points = {}
            self.archiveTable.setRowCount(0)
            self.toolBox.setItemText(0, self.tr('Archives'))
            self._toggle_all_buttons(enabled=False)

        self.archiveNameTemplate.setText(profile.new_archive_name)
        self.prunePrefixTemplate.setText(profile.prune_prefix)

        # Block signals so loading from the DB doesn't re-fire save_prune_setting
        # and overwrite prune_keep_within with stale QLineEdit text (#2493).
        self.prune_keep_within.blockSignals(True)
        try:
            self.prune_keep_within.setText(profile.prune_keep_within)
            for i in self.prune_intervals:
                widget = getattr(self, f'prune_{i}')
                widget.blockSignals(True)
                try:
                    widget.setValue(getattr(profile, f'prune_{i}'))
                finally:
                    widget.blockSignals(False)
        finally:
            self.prune_keep_within.blockSignals(False)

    def on_selection_change(self, selected=None, deselected=None):
        """
        React to a change of the selection of the archiveTableView.

        Enables or disables archive actions and the diff button.
        Makes sure at maximum 2 rows are selected.

        Parameters
        ----------
        selected : QItemSelection, optional
            The new selection.
        deselected : QItemSelection, optional
            The previous selection.
        """
        # handle selection of more than 2 rows
        selectionModel: QItemSelectionModel = self.archiveTable.selectionModel()
        indexes = selectionModel.selectedRows()
        # actions that are enabled only when a single archive is selected
        single_archive_action_buttons = [self.bMountArchive, self.bExtract, self.bRename]
        # actions that are enabled when at least one archive is selected
        multi_archive_action_buttons = [self.bDelete, self.bRefreshArchive]

        # Toggle archive actions frame
        layout: QLayout = self.fArchiveActions.layout()

        # task in progress -> disable all
        reason = ""
        if not self.repoactions_enabled:
            reason = self.tr("(borg already running)")

        # Disable the delete and refresh buttons if no archive is selected
        if self.repoactions_enabled and len(indexes) > 0:
            for button in multi_archive_action_buttons:
                button.setEnabled(True)
                button.setToolTip(self.tooltip_dict.get(button, ""))
        else:
            for button in multi_archive_action_buttons:
                button.setEnabled(False)
                button.setToolTip(self.tooltip_dict.get(button, "") + " " + self.tr("(Select minimum one archive)"))

        # Toggle diff button
        if self.repoactions_enabled and len(indexes) == 2:
            # Enable diff button
            self.bDiff.setEnabled(True)
            self.bDiff.setToolTip(self.tooltip_dict.get(self.bDiff, ""))
        else:
            # disable diff button
            self.bDiff.setEnabled(False)

            tooltip = self.tooltip_dict[self.bDiff]
            self.bDiff.setToolTip(tooltip + " " + reason or self.tr("(Select two archives)"))

        if self.repoactions_enabled and len(indexes) == 1:
            # Enable archive actions
            for widget in single_archive_action_buttons:
                widget.setEnabled(True)

            for index in range(layout.count()):
                widget = layout.itemAt(index).widget()
                if widget is not None:
                    widget.setToolTip(self.tooltip_dict.get(widget, ""))

            # refresh bMountArchive for the selected archive
            self.archive_mount.bmountarchive_refresh()
        else:
            reason = reason or self.tr("(Select exactly one archive)")

            # too few or too many selected.
            for widget in single_archive_action_buttons:
                tooltip = widget.toolTip()
                tooltip = self.tooltip_dict.setdefault(widget, tooltip)
                widget.setToolTip(tooltip + " " + reason)
                widget.setEnabled(False)

            # special treatment for dynamic mount/unmount button.
            self.archive_mount.bmountarchive_refresh()
            tooltip = self.bMountArchive.toolTip()
            self.bMountArchive.setToolTip(tooltip + " " + reason)

    def archive_copy(self, index=None):
        """
        Copy an archive name to the clipboard.

        Copies the first selected archive if no index is specified.
        """
        if index is None:
            indexes = self.archiveTable.selectionModel().selectedRows()

            if not indexes:
                return

            index = indexes[0]

        archive_name = self.archiveTable.item(index.row(), 4).text()

        data = QMimeData()
        data.setText(archive_name)

        QApplication.clipboard().setMimeData(data)

    def save_archive_template(self, tpl, key):
        profile = self.profile()
        try:
            preview = self.tr('Preview: %s') % format_archive_name(profile, tpl)
            setattr(profile, key, tpl)
            profile.save()
        except Exception:
            preview = self.tr('Error in archive name template.')

        if key == 'new_archive_name':
            self.archiveNamePreview.setText(preview)
        else:
            self.prunePrefixPreview.setText(preview)

    def check_action(self):
        params = BorgCheckJob.prepare(self.profile())
        if not params['ok']:
            self._set_status(params['message'])
            return

        # Conditions are met (borg binary available, etc)
        row_selected = self.archiveTable.selectionModel().selectedRows()
        if row_selected:
            archive_cell = self.archiveTable.item(row_selected[0].row(), 4)
            if archive_cell:
                archive_name = archive_cell.text()
                params['cmd'][-1] += f'::{archive_name}'

        job = BorgCheckJob(params['cmd'], params, self.profile().repo.id)
        job.updated.connect(self._set_status)
        job.result.connect(self.check_result)
        self._toggle_all_buttons(False)
        self.app.jobs_manager.add_job(job)

    def check_result(self, result):
        if result['returncode'] == 0:
            self._toggle_all_buttons(True)

    def compact_action(self):
        params = BorgCompactJob.prepare(self.profile())
        if params['ok']:
            job = BorgCompactJob(params['cmd'], params, self.profile().repo.id)
            job.updated.connect(self._set_status)
            job.result.connect(self.compact_result)
            self._toggle_all_buttons(False)
            self.app.jobs_manager.add_job(job)
        else:
            self._set_status(params['message'])

    def compact_result(self, result):
        self._toggle_all_buttons(True)

    def prune_action(self):
        params = BorgPruneJob.prepare(self.profile())
        if params['ok']:
            job = BorgPruneJob(params['cmd'], params, self.profile().repo.id)
            job.updated.connect(self._set_status)
            job.result.connect(self.prune_result)
            self._toggle_all_buttons(False)
            self.app.jobs_manager.add_job(job)
        else:
            self._set_status(params['message'])

    def prune_result(self, result):
        if result['returncode'] == 0:
            self._set_status(self.tr('Pruning finished.'))
            self.refresh_archive_list()
        else:
            self._toggle_all_buttons(True)

    def refresh_archive_list(self):
        params = BorgListRepoJob.prepare(self.profile())
        if params['ok']:
            job = BorgListRepoJob(params['cmd'], params, self.profile().repo.id)
            job.updated.connect(self._set_status)
            job.result.connect(self.list_result)
            self._toggle_all_buttons(False)
            self.app.jobs_manager.add_job(job)
        else:
            self._set_status(params['message'])

    def list_result(self, result):
        self._toggle_all_buttons(True)
        if result['returncode'] == 0:
            self._set_status(self.tr('Refreshed archives.'))
            self.populate_from_profile()

    def refresh_archive_info(self):
        selected_archives = self.archiveTable.selectionModel().selectedRows()

        archive_names = []
        for index in selected_archives:
            archive_names.append(self.archiveTable.item(index.row(), 4).text())

        self.remaining_refresh_archives = len(archive_names)  # number of archives to refresh
        self._toggle_all_buttons(False)
        for archive_name in archive_names:
            if archive_name is not None:
                params = BorgInfoArchiveJob.prepare(self.profile(), archive_name)
                if params['ok']:
                    job = BorgInfoArchiveJob(params['cmd'], params, self.profile().repo.id)
                    job.updated.connect(self._set_status)
                    job.result.connect(self.info_result)
                    self.app.jobs_manager.add_job(job)
                else:
                    self._set_status(params['message'])
                    return

    def info_result(self, result):
        self.remaining_refresh_archives -= 1
        if result['returncode'] == 0 and self.remaining_refresh_archives == 0:
            self._toggle_all_buttons(True)
            self._set_status(self.tr('Refreshed archives.'))
            self.populate_from_profile()

    def selected_archive_name(self):
        row_selected = self.archiveTable.selectionModel().selectedRows()
        if row_selected:
            archive_cell = self.archiveTable.item(row_selected[0].row(), 4)
            if archive_cell:
                return archive_cell.text()
        return None

    def save_prune_setting(self, new_value=None):
        profile = self.profile()
        for i in self.prune_intervals:
            setattr(profile, f'prune_{i}', getattr(self, f'prune_{i}').value())
        profile.prune_keep_within = self.prune_keep_within.text()
        profile.save()

    def cell_double_clicked(self, row=None, column=None):
        if not self.bRename.isEnabled():
            return

        if not row or not column:
            row = self.archiveTable.currentRow()
            column = self.archiveTable.currentColumn()

        if column == 3:
            archive_name = self.selected_archive_name()
            if not archive_name:
                return

            mount_point = self.mount_points.get(archive_name)

            if mount_point is not None:
                QDesktopServices.openUrl(QtCore.QUrl(f'file:///{mount_point}'))

        if column == 4:
            item = self.archiveTable.item(row, column)
            self.renamed_archive_original_name = item.text()
            self.is_editing = True
            item.setFlags(item.flags() | QtCore.Qt.ItemFlag.ItemIsEditable)
            self.archiveTable.editItem(item)

    def on_editing_finished(self, editor, hint):
        """Called when the user finishes editing a cell (By pressing Enter or clicking away)."""
        if not self.is_editing:
            return

        row = self.archiveTable.currentRow()
        column = self.archiveTable.currentColumn()

        self.cell_changed(row, column)

    def cell_changed(self, row, column):
        # return if this is not a name change
        if column != 4 or not self.is_editing:
            return

        item = self.archiveTable.item(row, column)
        new_name = item.text()
        profile = self.profile()

        # if the name hasn't changed or if this slot is called when first repopulating the table, do nothing.
        if new_name == self.renamed_archive_original_name:
            self.is_editing = False
            return

        # if new name is blank, revert to the original name
        if not new_name:
            item.setText(self.renamed_archive_original_name)
            self._set_status(self.tr('Archive name cannot be blank.'))
            self.is_editing = False
            return

        # check if the new name already exists
        new_name_exists = ArchiveModel.get_or_none(name=new_name, repo=profile.repo)
        if new_name_exists is not None:
            self._set_status(self.tr('An archive with this name already exists.'))
            item.setText(self.renamed_archive_original_name)
            self.is_editing = False
            return

        params = BorgRenameJob.prepare(profile, self.renamed_archive_original_name, new_name)
        if not params['ok']:
            self._set_status(params['message'])
            self.is_editing = False
            return

        self._set_status(self.tr('Renaming archive...'))
        job = BorgRenameJob(params['cmd'], params, self.profile().repo.id)
        job.updated.connect(self._set_status)
        job.result.connect(self.rename_result)
        self._toggle_all_buttons(False)
        self.app.jobs_manager.add_job(job)
        self.is_editing = False

    def row_of_archive(self, archive_name):
        items = self.archiveTable.findItems(archive_name, QtCore.Qt.MatchFlag.MatchExactly)
        rows = [item.row() for item in items if item.column() == 4]
        return rows[0] if rows else None

    def confirm_dialog(self, title, text):
        msg = QMessageBox()
        msg.setIcon(QMessageBox.Icon.Information)
        msg.setText(text)
        msg.setWindowTitle(title)
        msg.setStandardButtons(QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.Cancel)
        msg.button(QMessageBox.StandardButton.Yes).setText(self.tr("Yes"))
        msg.button(QMessageBox.StandardButton.Cancel).setText(self.tr("Cancel"))
        return msg.exec() == QMessageBox.StandardButton.Yes

    def delete_action(self):
        # Since this function modify the UI, we can't put the whole function in a JobQueue.

        # determine selected archives
        archives = []
        for index in self.archiveTable.selectionModel().selectedRows():
            archive_cell = self.archiveTable.item(index.row(), 4)
            if archive_cell:
                archives.append(archive_cell.text())

        if not archives:
            self._set_status(self.tr("No archive selected"))
            return

        params = BorgDeleteJob.prepare(self.profile(), archives)
        if not params['ok']:
            self._set_status(params['message'])
            return

        if len(archives) > 1:
            body = self.tr("Are you sure you want to delete all the selected archives?")
        else:
            body = self.tr("Are you sure you want to delete the selected archive?")
        if not self.confirm_dialog(self.tr("Confirm deletion"), body):
            return

        job = BorgDeleteJob(params['cmd'], params, self.profile().repo.id)
        job.updated.connect(self._set_status)
        job.result.connect(self.delete_result)
        self._toggle_all_buttons(False)
        self.app.jobs_manager.add_job(job)

    def delete_result(self, result):
        archives = result['params']['archives']
        if result['returncode'] == 0:
            if len(archives) > 1:
                status = self.tr('Archives deleted.')
            else:
                status = self.tr('Archive deleted.')
            self._set_status(status)

            # remove rows from list and database
            for archive in archives:
                for entry in self.archiveTable.findItems(archive, QtCore.Qt.MatchFlag.MatchExactly):
                    self.archiveTable.removeRow(entry.row())
                archive_obj = ArchiveModel.get_or_none(name=archive)
                if archive_obj:
                    archive_obj.delete_instance()

        self._toggle_all_buttons(True)

    def diff_action(self):
        """
        Handle the diff button being clicked.

        Exactly two archives must be selected in `archiveTable`. This is
        usually enforced by `on_selection_change`.
        """
        selected_archives = self.archiveTable.selectionModel().selectedRows()
        profile = self.profile()

        name1 = self.archiveTable.item(selected_archives[0].row(), 4).text()
        name2 = self.archiveTable.item(selected_archives[1].row(), 4).text()

        archive1, archive2 = (
            profile.repo.archives.select()
            .where((ArchiveModel.name == name1) | (ArchiveModel.name == name2))
            .order_by(ArchiveModel.time.desc())
        )

        archive_name_newer = archive1.name
        archive_name_older = archive2.name

        # Start diff job
        params = BorgDiffJob.prepare(profile, archive_name_older, archive_name_newer)

        if params['ok']:
            self._toggle_all_buttons(False)
            job = BorgDiffJob(params['cmd'], params, self.profile().repo.id)
            job.updated.connect(self.mountErrors.setText)
            job.result.connect(self.list_diff_result)
            self.app.jobs_manager.add_job(job)
        else:
            self._set_status(params['message'])

    def list_diff_result(self, result):
        """
        Process the result of the `BorgDiffJob`.

        The `BorgDiffJob` was initiated by `diff_action`.

        Parameters
        ----------
        result : dict
            The BorgJob result.
        """
        self._set_status('')
        if result['returncode'] == 0:
            archive_newer = ArchiveModel.get(name=result['params']['archive_name_newer'])
            archive_older = ArchiveModel.get(name=result['params']['archive_name_older'])
            self._set_status(self.tr("Processing diff results."))

            model = DiffTree()

            self._t = diff_result.ParseThread(result['data'], result['params']['json_lines'], model)
            self._t.finished.connect(lambda: self.show_diff_result(archive_newer, archive_older, model))
            self._t.start()

    def show_diff_result(self, archive_newer, archive_older, model):
        self._t = None

        # show dialog
        self._toggle_all_buttons(True)
        self._set_status('')
        window = DiffResultDialog(archive_newer, archive_older, model)
        window.setParent(self)
        window.setWindowFlags(Qt.WindowType.Window)
        window.setWindowModality(Qt.WindowModality.NonModal)
        self._resultwindow = window  # for testing
        window.show()

    def rename_result(self, result):
        if result['returncode'] == 0:
            self.refresh_archive_info()
            self._set_status(self.tr('Archive renamed.'))
            self.renamed_archive_original_name = None
            self.populate_from_profile()
        else:
            self._toggle_all_buttons(True)

    def toggle_compact_button_visibility(self):
        """
        Enable or disable the compact button depending on the Borg version.
        This function runs once on startup, and every time the profile is changed.
        """
        if borg_compat.check("COMPACT_SUBCOMMAND"):
            self.compactButton.setEnabled(True)
            self.compactButton.setToolTip(self.tooltip_dict[self.compactButton])
        else:
            self.compactButton.setEnabled(False)
            tooltip = self.tooltip_dict[self.compactButton]
            self.compactButton.setToolTip(tooltip + " " + self.tr("(This feature needs Borg 1.2.0 or higher)"))



from vorta.store.models import ArchiveModel, RepoModel
from vorta.utils import borg_compat

from .borg_job import BorgJob


class BorgRenameJob(BorgJob):
    def started_event(self):
        self.app.backup_started_event.emit()
        self.app.backup_progress_event.emit(f"[{self.params['profile_name']}] {self.tr('Renaming archive…')}")

    def log_event(self, msg):
        self.app.backup_log_event.emit(msg)

    @classmethod
    def prepare(cls, profile, old_archive_name, new_archive_name):
        ret = super().prepare(profile)
        if not ret['ok']:
            return ret
        else:
            ret['ok'] = False  # Set back to false, so we can do our own checks here.

        cmd = ['borg', 'rename', '--info', '--log-json']
        if borg_compat.check('V2'):
            cmd.extend(["-r", profile.repo.url, old_archive_name, new_archive_name])
        else:
            cmd.extend([f'{profile.repo.url}::{old_archive_name}', new_archive_name])

        ret['old_archive_name'] = old_archive_name
        ret['new_archive_name'] = new_archive_name
        ret['repo_url'] = profile.repo.url
        ret['ok'] = True
        ret['cmd'] = cmd

        return ret

    def process_result(self, result):
        if result['returncode'] == 0:
            repo = RepoModel.get(url=result['params']['repo_url'])
            renamed_archive = ArchiveModel.get(name=result['params']['old_archive_name'], repo=repo)
            renamed_archive.name = result['params']['new_archive_name']
            renamed_archive.save()

            self.app.backup_progress_event.emit(f"[{self.params['profile_name']}] {self.tr('Archive renamed.')}")



from __future__ import annotations

import logging
from collections.abc import Mapping
from datetime import datetime
from enum import Enum
from typing import Any, NamedTuple

from PyQt6 import QtDBus
from PyQt6.QtCore import QObject, QVersionNumber, pyqtSignal, pyqtSlot

from vorta.network_status.abc import NetworkStatusMonitor, SystemWifiInfo

logger = logging.getLogger(__name__)


class NetworkManagerMonitor(NetworkStatusMonitor):
    def __init__(self, nm_adapter: NetworkManagerDBusAdapter | None = None) -> None:
        super().__init__()
        self._nm = nm_adapter or NetworkManagerDBusAdapter.get_system_nm_adapter()
        self._nm.network_status_changed.connect(self.network_status_changed)

    def is_network_metered(self) -> bool:
        try:
            return self._nm.get_global_metered_status() in (
                NMMetered.YES,
                NMMetered.GUESS_YES,
            )
        except DBusException:
            logger.exception("Failed to check if network is metered, assuming it isn't")
            return False

    def is_network_active(self) -> bool:
        try:
            return self._nm.is_network_connected()
        except DBusException:
            logger.exception("Failed to check connectivity state. Assuming connected")
            return True

    def get_current_wifi(self) -> str | None:
        # Check all active connections for WiFi, not just the primary one.
        # When connected via both WiFi and wired (e.g. docking station), the primary
        # connection may be wired, but the WiFi SSID is still needed for network-based
        # backup restrictions. See #1775.
        try:
            active_paths = self._nm.get_active_connections_paths()
            for active_connection_path in active_paths:
                try:
                    active_connection = self._nm.get_active_connection_info(active_connection_path)
                    if active_connection.type == '802-11-wireless':
                        settings = self._nm.get_settings(active_connection.connection)
                        ssid = self._get_ssid_from_settings(settings)
                        if ssid:
                            return ssid
                except DBusException:
                    logger.debug("Failed to get info for active connection %s", active_connection_path)
                    continue
        except DBusException:
            logger.exception("Failed to get currently connected WiFi network, assuming none")
        return None

    def get_known_wifis(self) -> list[SystemWifiInfo]:
        wifis: list[SystemWifiInfo] = []
        try:
            connections_paths = self._nm.get_connections_paths()
        except DBusException:
            logger.exception("Failed to list connections")
            return wifis

        for connection_path in connections_paths:
            try:
                settings = self._nm.get_settings(connection_path)
            except DBusException:
                logger.warning("Couldn't load settings for %s", connection_path, exc_info=True)
            else:
                ssid = self._get_ssid_from_settings(settings)
                if ssid:
                    timestamp = settings['connection'].get('timestamp')
                    wifis.append(
                        SystemWifiInfo(
                            ssid=ssid,
                            last_connected=timestamp and datetime.utcfromtimestamp(timestamp),
                        )
                    )
        return wifis

    def _get_ssid_from_settings(self, settings: Mapping[str, Mapping[str, Any]]) -> str | None:
        wireless_settings = settings.get('802-11-wireless') or {}
        raw_ssid = wireless_settings.get('ssid')
        ssid = raw_ssid and decode_ssid(raw_ssid)
        return ssid


def decode_ssid(raw_ssid: list[int]) -> str | None:
    """SSIDs are binary strings, but we need something to show to the user."""
    # Best effort UTF-8 decoding, as most SSIDs are UTF-8 (or even ASCII)
    str_ssid = bytes(raw_ssid).decode('utf-8', 'surrogateescape')
    if str_ssid.isprintable():
        return str_ssid
    else:
        return ''.join(c if c.isprintable() else ascii(c)[1:-1] for c in str_ssid)


class UnsupportedException(Exception):
    """NetworkManager is not available"""


class DBusException(Exception):
    """Failed to call a DBus method"""


class NetworkManagerDBusAdapter(QObject):
    """Simple adapter to NetworkManager's DBus interface.
    This should be the only part of NM support that needs manual testing."""

    BUS_NAME = 'org.freedesktop.NetworkManager'
    NM_PATH = '/org/freedesktop/NetworkManager'
    INTERFACE_NAME = 'org.freedesktop.NetworkManager'
    # Use the NMState everywhere in lieu of Connected. There is no change signal for
    # Connected and it appears that the connected state changes after the state change.
    # i.e. immediately asking for current connectivity can return the old value
    SIGNAL_NAME = 'StateChanged'

    network_status_changed = pyqtSignal(bool, name="networkStatusChanged")

    def __init__(self, parent: QObject | None, bus: QtDBus.QDBusConnection) -> None:
        super().__init__(parent)
        self._bus = bus
        self._bus.connect(
            self.BUS_NAME, self.NM_PATH, self.INTERFACE_NAME, self.SIGNAL_NAME, 'u', self.networkStateChanged
        )
        self._nm = self._get_iface(self.NM_PATH, 'org.freedesktop.NetworkManager')

    @classmethod
    def get_system_nm_adapter(cls) -> NetworkManagerDBusAdapter:
        bus = QtDBus.QDBusConnection.systemBus()
        if not bus.isConnected():
            raise UnsupportedException("Can't connect to system bus")
        nm_adapter = cls(parent=None, bus=bus)
        if not nm_adapter.isValid():
            raise UnsupportedException("Can't connect to NetworkManager")
        return nm_adapter

    @pyqtSlot("unsigned int")
    def networkStateChanged(self, state: int) -> None:
        logger.debug(f'network state changed: {state}')
        # https://www.networkmanager.dev/docs/api/latest/nm-dbus-types.html#NMState
        self.network_status_changed.emit(_is_network_connected(NMState(state)))

    def isValid(self) -> bool:
        if not self._nm.isValid():
            return False
        nm_version = self._get_nm_version()
        if nm_version < QVersionNumber(1, 2):
            logger.warning(
                'NetworkManager version 1.2 or later required, found %s',
                nm_version.toString(),
            )
            return False
        return True

    def is_network_connected(self) -> bool:
        return _is_network_connected(self.get_network_state())

    def get_network_state(self) -> NMState:
        return NMState(read_dbus_property(self._nm, 'State'))

    def get_primary_connection_path(self) -> str | None:
        return read_dbus_property(self._nm, 'PrimaryConnection')

    def get_active_connections_paths(self) -> list[str]:
        return read_dbus_property(self._nm, 'ActiveConnections') or []

    def get_active_connection_info(self, active_connection_path: str) -> ActiveConnectionInfo:
        active_connection = self._get_iface(active_connection_path, 'org.freedesktop.NetworkManager.Connection.Active')
        return ActiveConnectionInfo(
            connection=read_dbus_property(active_connection, 'Connection'),
            type=read_dbus_property(active_connection, 'Type'),
        )

    def get_connections_paths(self) -> list[str]:
        settings_manager = self._get_iface(self.NM_PATH + '/Settings', 'org.freedesktop.NetworkManager.Settings')
        return get_result(settings_manager.call('ListConnections'))

    def get_settings(self, connection_path: str) -> Mapping[str, Mapping[str, Any]]:
        settings = self._get_iface(connection_path, 'org.freedesktop.NetworkManager.Settings.Connection')
        return get_result(settings.call('GetSettings'))

    def get_global_metered_status(self) -> NMMetered:
        return NMMetered(read_dbus_property(self._nm, 'Metered'))

    def _get_nm_version(self) -> QVersionNumber:
        version, _suffindex = QVersionNumber.fromString(read_dbus_property(self._nm, 'Version'))
        return version

    def _get_iface(self, path: str, interface: str) -> QtDBus.QDBusInterface:
        return QtDBus.QDBusInterface(self.BUS_NAME, path, interface, self._bus)


def _is_network_connected(state: NMState) -> bool:
    # We treat site and global as connected because having a default route means you
    # can reach something. This might need to include LOCAL eventually depending on use
    # cases
    return state in (
        NMState.NM_STATE_CONNECTED_SITE,
        NMState.NM_STATE_CONNECTED_GLOBAL,
    )


def read_dbus_property(obj: QtDBus.QDBusInterface, property: str) -> Any:
    # QDBusInterface.property() didn't work for some reason
    props = QtDBus.QDBusInterface(obj.service(), obj.path(), 'org.freedesktop.DBus.Properties', obj.connection())
    msg = props.call('Get', obj.interface(), property)
    return get_result(msg)


def get_result(msg: QtDBus.QDBusMessage) -> Any:
    if msg.type() == msg.MessageType.ReplyMessage:
        return msg.arguments()[0]
    else:
        raise DBusException("DBus call failed: {}".format(msg.arguments()))


class ActiveConnectionInfo(NamedTuple):
    connection: str
    type: str


class NMMetered(Enum):
    UNKNOWN = 0
    YES = 1
    NO = 2
    GUESS_YES = 3
    GUESS_NO = 4


class NMDeviceType(Enum):
    # Only the types we care about
    UNKNOWN = 0
    WIFI = 2


class NMState(Enum):
    """https://www.networkmanager.dev/docs/api/latest/nm-dbus-types.html#NMState"""

    NM_STATE_UNKNOWN = 0
    NM_STATE_DISABLED = 10
    NM_STATE_DISCONNECTED = 20
    NM_STATE_DISCONNECTING = 30
    NM_STATE_CONNECTING = 40
    NM_STATE_CONNECTED_LOCAL = 50
    NM_STATE_CONNECTED_SITE = 60
    NM_STATE_CONNECTED_GLOBAL = 70



import os
import sys

from PyQt6.QtGui import QIcon, QImage, QPixmap
from PyQt6.QtWidgets import QApplication, QMenu, QSystemTrayIcon

from vorta.store.models import BackupProfileModel
from vorta.utils import get_asset


class TrayMenu(QSystemTrayIcon):
    def __init__(self, parent: QApplication = None):
        QSystemTrayIcon.__init__(self, parent)
        self.app = parent
        self.set_tray_icon()
        menu = QMenu()

        # Workaround to get `activated` signal on Unity: https://stackoverflow.com/a/43683895/3983708
        menu.aboutToShow.connect(self.build_menu)

        self.setContextMenu(menu)
        self.build_menu()

        self.activated.connect(self.on_activation)
        self._palette_connection = self.app.paletteChanged.connect(lambda p: self.set_tray_icon())
        self.destroyed.connect(self._on_destroyed)
        self.setVisible(True)
        self.show()

    def on_activation(self, reason):
        """
        XDG-compliant tray behavior: right click to open a context menu, and left click to
        restore the application.

        If XDG_CURRENT_DESKTOP isn't set, always open the tray menu (macOS)
        """
        if reason in [
            QSystemTrayIcon.ActivationReason.Trigger,
            QSystemTrayIcon.ActivationReason.DoubleClick,
        ] and os.environ.get('XDG_CURRENT_DESKTOP'):
            self.app.toggle_main_window_visibility()

    def build_menu(self):
        """Build system tray menu based on current state."""

        menu = self.contextMenu()
        menu.clear()

        open_action = menu.addAction(self.tr('Vorta for Borg Backup'))
        open_action.triggered.connect(self.app.open_main_window_action)

        menu.addSeparator()

        next_task_time = self.app.scheduler.next_job()
        status = menu.addAction(next_task_time)
        status.setEnabled(False)

        if self.app.jobs_manager.is_worker_running():
            status.setText(self.tr('Task in progress'))
            cancel_action = menu.addAction(self.tr('Cancel Backup'))
            cancel_action.triggered.connect(self.app.backup_cancelled_event.emit)
        else:
            status.setText(self.tr('Next Task: %s') % next_task_time)
            profiles = BackupProfileModel.select()
            if profiles.count() > 1:
                profile_menu = menu.addMenu(self.tr('Backup Now'))
                for profile in sorted(profiles, key=lambda p: (p.name.casefold(), p.name)):
                    new_item = profile_menu.addAction(profile.name)
                    new_item.triggered.connect(lambda state, i=profile.id: self.app.create_backup_action(i))
            else:
                profile = profiles.first()
                profile_menu = menu.addAction(self.tr('Backup Now'))
                profile_menu.triggered.connect(lambda state, i=profile.id: self.app.create_backup_action(i))

        menu.addSeparator()

        exit_action = menu.addAction(self.tr('Quit'))
        exit_action.triggered.connect(self.app.quit)

    def set_tray_icon(self, active=False):
        """
        Use white tray icon, when on Gnome or in dark mode.
        Use mask tray icon when on macOS.
        Otherwise use dark icon.
        """
        if sys.platform == 'darwin':
            icon_name = f"icons/hdd-o{'-active' if active else ''}-mask.svg"
            icon = QIcon(QPixmap(QImage.fromData(open(get_asset(icon_name), 'rb').read()).scaledToHeight(128)))
            icon.setIsMask(True)
        else:
            icon_name = f"icons/hdd-o{'-active' if active else ''}.png"
            icon = QIcon(get_asset(icon_name))
        self.setIcon(icon)

    def _on_destroyed(self):
        try:
            self.app.paletteChanged.disconnect(self._palette_connection)
        except (TypeError, RuntimeError):
            pass



from PyQt6 import uic
from PyQt6.QtWidgets import (
    QAbstractItemView,
    QHeaderView,
    QTableWidgetItem,
)

from vorta import config
from vorta.i18n.richtext import format_richtext, link
from vorta.store.models import EventLogModel
from vorta.utils import get_asset
from vorta.views.base_tab import BaseTab

uifile = get_asset('UI/log_page.ui')
LogTableUI, LogTableBase = uic.loadUiType(uifile)


class LogTableColumn:
    Time = 0
    Category = 1
    Subcommand = 2
    Repository = 3
    ReturnCode = 4


class LogPage(BaseTab, LogTableBase, LogTableUI):
    def __init__(self, parent=None, profile_provider=None):
        super().__init__(parent=parent, profile_provider=profile_provider)
        self.setupUi(self)
        self.init_ui()
        self.track_profile_change(self.populate_logs, call_now=True)
        self.track_backup_finished(self.populate_logs)

    def init_ui(self):
        self.logPage.setAlternatingRowColors(True)
        header = self.logPage.horizontalHeader()
        header.setVisible(True)
        [header.setSectionResizeMode(i, QHeaderView.ResizeMode.ResizeToContents) for i in range(5)]
        header.setSectionResizeMode(3, QHeaderView.ResizeMode.Stretch)
        self.logPage.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.logPage.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)

        template = self.logLink.text()
        log_link = link(f"file://{config.LOG_DIR}", self.tr('View the logs'))
        self.logLink.setText(format_richtext(template, log_link))

    def populate_logs(self):
        profile = self.profile()
        event_logs = [
            s
            for s in EventLogModel.select()
            .where(EventLogModel.profile == profile.id)
            .order_by(EventLogModel.start_time.desc())
        ]

        sorting = self.logPage.isSortingEnabled()
        self.logPage.setSortingEnabled(False)
        self.logPage.setRowCount(len(event_logs))
        for row, log_line in enumerate(event_logs):
            formatted_time = log_line.start_time.strftime('%Y-%m-%d %H:%M')
            self.logPage.setItem(row, LogTableColumn.Time, QTableWidgetItem(formatted_time))
            self.logPage.setItem(row, LogTableColumn.Category, QTableWidgetItem(log_line.category))
            self.logPage.setItem(row, LogTableColumn.Subcommand, QTableWidgetItem(log_line.subcommand))
            self.logPage.setItem(row, LogTableColumn.Repository, QTableWidgetItem(log_line.repo_url))
            self.logPage.setItem(row, LogTableColumn.ReturnCode, QTableWidgetItem(str(log_line.returncode)))
        self.logPage.setSortingEnabled(sorting)



"""
This module provides the app's data store using Peewee with SQLite.

At the bottom there is a simple schema migration system.
"""

from __future__ import annotations

import json
import logging
from datetime import datetime
from enum import Enum
from typing import Any

import peewee as pw
from playhouse import signals

from vorta.utils import slugify
from vorta.views.utils import get_exclusion_presets

DB = pw.Proxy()
logger = logging.getLogger(__name__)


class JSONField(pw.TextField):
    """
    Class to "fake" a JSON field with a text field. Not efficient but works nicely.

    From: https://gist.github.com/rosscdh/f4f26758b0228f475b132c688f15af2b
    """

    def db_value(self, value: Any) -> str | None:
        """Convert the python value for storage in the database."""
        return value if value is None else json.dumps(value)

    def python_value(self, value: str | None) -> Any:
        """Convert the database value to a pythonic value."""
        return value if value is None else json.loads(value)


class BaseModel(signals.Model):
    """Common model superclass."""


class RepoModel(BaseModel):
    """A single remote repo with unique URL."""

    url = pw.CharField(unique=True)
    name = pw.CharField(default='')
    added_at = pw.DateTimeField(default=datetime.now)
    encryption = pw.CharField(null=True)
    unique_size = pw.IntegerField(null=True)
    unique_csize = pw.IntegerField(null=True)
    total_size = pw.IntegerField(null=True)
    total_unique_chunks = pw.IntegerField(null=True)
    create_backup_cmd = pw.CharField(default='')
    extra_borg_arguments = pw.CharField(default='')

    def is_remote_repo(self) -> bool:
        return not self.url.startswith('/')

    def is_shared_with_other_profiles(self, excluding_profile_id: int | None = None) -> bool:
        """Return whether another profile still references this repository."""
        query = BackupProfileModel.select().where(BackupProfileModel.repo == self)
        if excluding_profile_id is not None:
            query = query.where(BackupProfileModel.id != excluding_profile_id)  # type: ignore[attr-defined]
        return query.exists()

    class Meta:
        database = DB


class RepoPassword(BaseModel):
    """Fallback to save repo passwords. Only used if no Keyring available."""

    url = pw.CharField(unique=True)
    password = pw.CharField()

    class Meta:
        database = DB


class BackupProfileModel(BaseModel):
    """Allows the user to switch between different configurations."""

    name = pw.CharField()
    added_at = pw.DateTimeField(default=datetime.now)
    repo = pw.ForeignKeyField(RepoModel, default=None, null=True)
    ssh_key = pw.CharField(default=None, null=True)
    compression = pw.CharField(default='lz4')
    exclude_patterns = pw.TextField(null=True)
    exclude_if_present = pw.TextField(null=True, default="[] .nobackup\n[] CACHEDIR.TAG")
    schedule_mode = pw.CharField(default='off')
    schedule_interval_count = pw.IntegerField(default=3)
    schedule_interval_unit = pw.CharField(default='hours')
    schedule_fixed_hour = pw.IntegerField(default=3)
    schedule_fixed_minute = pw.IntegerField(default=42)
    schedule_interval_hours = pw.IntegerField(default=3)  # no longer used
    schedule_interval_minutes = pw.IntegerField(default=42)  # no longer used
    schedule_make_up_missed = pw.BooleanField(default=True)
    validation_on = pw.BooleanField(default=True)
    validation_weeks = pw.IntegerField(default=3)
    compaction_on = pw.BooleanField(default=False)
    compaction_weeks = pw.IntegerField(default=3)
    prune_on = pw.BooleanField(default=False)
    prune_hour = pw.IntegerField(default=2)
    prune_day = pw.IntegerField(default=7)
    prune_week = pw.IntegerField(default=4)
    prune_month = pw.IntegerField(default=6)
    prune_year = pw.IntegerField(default=2)
    prune_keep_within = pw.CharField(default='10H', null=True)
    new_archive_name = pw.CharField(default="{hostname}-{now:%Y-%m-%d-%H%M%S}")
    prune_prefix = pw.CharField(default="{hostname}-")
    pre_backup_cmd = pw.CharField(default='')
    post_backup_cmd = pw.CharField(default='')
    dont_run_on_metered_networks = pw.BooleanField(default=True)

    def refresh(self) -> BackupProfileModel:
        return type(self).get(self._pk_expr())  # type: ignore[attr-defined]

    def slug(self) -> str:
        return slugify(self.name)

    def get_combined_exclusion_string(self) -> str:
        allPresets = get_exclusion_presets()
        excludes = ""

        if (
            ExclusionModel.select()
            .where(
                ExclusionModel.profile == self,
                ExclusionModel.enabled,
                ExclusionModel.source == ExclusionModel.SourceFieldOptions.CUSTOM.value,
            )
            .count()
            > 0
        ):
            excludes = "# custom added rules\n"

        for exclude in ExclusionModel.select().where(
            ExclusionModel.profile == self,
            ExclusionModel.enabled,
            ExclusionModel.source == ExclusionModel.SourceFieldOptions.CUSTOM.value,
        ):
            excludes += f"{exclude.name}\n"

        raw_excludes = self.exclude_patterns
        if raw_excludes:
            excludes += "\n# raw exclusions\n"
            excludes += raw_excludes
            excludes += "\n"

        # go through all source=='preset' exclusions, find the name in the allPresets dict, and add the patterns
        for exclude in ExclusionModel.select().where(
            ExclusionModel.profile == self,
            ExclusionModel.enabled,
            ExclusionModel.source == ExclusionModel.SourceFieldOptions.PRESET.value,
        ):
            if exclude.name not in allPresets:
                logger.warning("Exclusion preset %s not found in built-in presets.", exclude.name)
                continue
            excludes += f"\n# {exclude.name}\n"
            for pattern in allPresets[exclude.name]['patterns']:
                excludes += f"{pattern}\n"

        return excludes

    class Meta:
        database = DB


class ExclusionModel(BaseModel):
    """
    If this is a user created exclusion, the name will be the same as the pattern added. For exclusions added from
    presets, the name will be the same as the preset name. Duplicate patterns are already handled by Borg.
    """

    class SourceFieldOptions(Enum):
        CUSTOM = 'custom'
        PRESET = 'preset'

    profile = pw.ForeignKeyField(BackupProfileModel, backref='exclusions')
    name = pw.CharField()
    enabled = pw.BooleanField(default=True)
    source = pw.CharField(default=SourceFieldOptions.CUSTOM.value)

    class Meta:
        database = DB


class SourceFileModel(BaseModel):
    """A folder to be backed up, related to a Backup Configuration."""

    dir = pw.CharField()
    dir_size = pw.BigIntegerField(default=-1)
    dir_files_count = pw.BigIntegerField(default=-1)
    path_isdir = pw.BooleanField(default=False)
    profile = pw.ForeignKeyField(BackupProfileModel, default=1)
    added_at = pw.DateTimeField(default=datetime.utcnow)

    class Meta:
        database = DB
        table_name = 'sourcedirmodel'


class ArchiveModel(BaseModel):
    """An archive in a remote repository."""

    snapshot_id = pw.CharField()
    name = pw.CharField()
    repo = pw.ForeignKeyField(RepoModel, backref='archives')
    time = pw.DateTimeField()
    duration = pw.FloatField(null=True)
    size = pw.IntegerField(null=True)
    trigger = pw.CharField(null=True)

    class Meta:
        database = DB


class WifiSettingModel(BaseModel):
    """Save Wifi Settings"""

    ssid = pw.CharField()
    last_connected = pw.DateTimeField(null=True)
    allowed = pw.BooleanField(default=True)
    profile = pw.ForeignKeyField(BackupProfileModel, default=1)

    class Meta:
        database = DB


class EventLogModel(BaseModel):
    """Keep a log of background jobs."""

    start_time = pw.DateTimeField(default=datetime.now)
    end_time = pw.DateTimeField(default=datetime.now)
    category = pw.CharField()
    subcommand = pw.CharField(null=True)
    message = pw.CharField(null=True)
    returncode = pw.IntegerField(default=-1)
    params = JSONField(null=True)
    profile = pw.CharField(null=True)
    repo_url = pw.CharField(null=True)

    class Meta:
        database = DB


class SchemaVersion(BaseModel):
    """Keep DB version to apply the correct migrations."""

    version = pw.IntegerField()
    changed_at = pw.DateTimeField(default=datetime.now)

    class Meta:
        database = DB


class SettingsModel(BaseModel):
    """App settings unrelated to a single profile or repo"""

    key = pw.CharField(unique=True)
    value = pw.BooleanField(default=False)
    str_value = pw.CharField(default='')
    label = pw.CharField()
    group = pw.CharField(default='')  # Settings group name and label
    tooltip = pw.CharField(default='')  # optional tooltip for `checkbox` type
    type = pw.CharField()

    class Meta:
        database = DB



import logging
import os

from vorta.store.models import SettingsModel
from vorta.utils import SHELL_PATTERN_ELEMENT, borg_compat

from .borg_job import BorgJob

logger = logging.getLogger(__name__)


class BorgMountJob(BorgJob):
    def started_event(self):
        self.updated.emit(self.tr('Mounting archive into folder…'))

    @classmethod
    def prepare(cls, profile, archive: str = None):
        ret = super().prepare(profile)
        if not ret['ok']:
            return ret
        else:
            ret['ok'] = False  # Set back to false, so we can do our own checks here.

        cmd = ['borg', '--log-json', 'mount']

        # Try to override existing permissions when mounting an archive. May help to read
        # files that come from a different system, like a restrictive NAS.
        override_mount_permissions = SettingsModel.get(key='override_mount_permissions').value
        if override_mount_permissions:
            cmd += ['-o', f"umask=0277,uid={os.getuid()}"]

        if borg_compat.check('V2'):
            cmd.extend(["-r", profile.repo.url])

            if archive:
                # in shell patterns ?, * and [...] have a special meaning
                pattern = SHELL_PATTERN_ELEMENT.sub(r'\\1', archive)  # escape them
                cmd.extend(['-a', pattern])
        else:
            source = f'{profile.repo.url}'

            if archive:
                source += f'::{archive}'

            cmd.append(source)

        if archive:
            ret['mounted_archive'] = archive

        ret['ok'] = True
        ret['cmd'] = cmd

        return ret



import logging
import sys
from pathlib import Path

from PyQt6 import QtCore, uic
from PyQt6.QtCore import QPoint, Qt
from PyQt6.QtCore import pyqtSignal as Signal
from PyQt6.QtGui import QFontMetrics, QKeySequence, QShortcut
from PyQt6.QtWidgets import (
    QApplication,
    QCheckBox,
    QFileDialog,
    QListWidgetItem,
    QMessageBox,
    QToolTip,
)

from vorta.profile_export import ImportFailedException, ProfileExport
from vorta.store.models import BackupProfileModel, SettingsModel
from vorta.utils import (
    borg_compat,
    get_asset,
    get_network_status_monitor,
    is_system_tray_available,
)
from vorta.views.dialogs.profile.export_window import ExportWindow
from vorta.views.dialogs.profile.import_window import ImportWindow
from vorta.views.dialogs.profile.profile_add_edit import AddProfileWindow, EditProfileWindow
from vorta.views.partials.loading_button import LoadingButton
from vorta.views.utils import get_colored_icon

from .about_tab import AboutTab
from .archive_tab import ArchiveTab
from .misc_tab import MiscTab
from .repo_tab import RepoTab
from .schedule_tab import ScheduleTab
from .source_tab import SourceTab

uifile = get_asset('UI/main_window.ui')
MainWindowUI, MainWindowBase = uic.loadUiType(uifile)

logger = logging.getLogger(__name__)


class MainWindow(MainWindowBase, MainWindowUI):
    loaded = Signal()

    def __init__(self, parent=None):
        super().__init__()
        self.setupUi(self)
        self.setWindowTitle('Vorta for Borg Backup')
        self.app = parent
        self.setWindowIcon(get_colored_icon("icon"))
        if sys.platform.startswith('linux'):
            self.app.setDesktopFileName('com.borgbase.Vorta')
        self.setWindowFlags(QtCore.Qt.WindowType.WindowCloseButtonHint | QtCore.Qt.WindowType.WindowMinimizeButtonHint)
        self.createStartBtn = LoadingButton(self.tr("Start Backup"))
        self.gridLayout.addWidget(self.createStartBtn, 0, 0, 1, 1)
        self.createStartBtn.setGif(get_asset("icons/loading"))

        # set log label height to two lines
        fontmetrics: QFontMetrics = self.logText.fontMetrics()
        self.logText.setMinimumHeight(fontmetrics.lineSpacing() * 2 + fontmetrics.leading())

        # Use previous window state
        previous_window_width = SettingsModel.get(key='previous_window_width')
        previous_window_height = SettingsModel.get(key='previous_window_height')
        self.resize(int(previous_window_width.str_value), int(previous_window_height.str_value))

        # Select previously used profile, if available
        prev_profile_id = SettingsModel.get(key='previous_profile_id')
        self.current_profile = BackupProfileModel.get_or_none(id=prev_profile_id.str_value)
        if self.current_profile is None:
            profiles = BackupProfileModel.select()
            self.current_profile = min(profiles, key=lambda p: (p.name.casefold(), p.name))

        # Load tab models
        self.repoTab = RepoTab(self.repoTabSlot, profile_provider=self.get_current_profile)
        self.sourceTab = SourceTab(self.sourceTabSlot, profile_provider=self.get_current_profile)
        self.archiveTab = ArchiveTab(self.archiveTabSlot, app=self.app, profile_provider=self.get_current_profile)
        self.scheduleTab = ScheduleTab(self.scheduleTabSlot, profile_provider=self.get_current_profile)
        self.miscTab = MiscTab(self.SettingsTabSlot, profile_provider=self.get_current_profile)
        self.aboutTab = AboutTab(self.AboutTabSlot, profile_provider=self.get_current_profile)
        self.aboutTab.set_borg_details(borg_compat.version, borg_compat.path)
        self.miscWidget.hide()
        self.tabWidget.setCurrentIndex(0)

        self.repoTab.repo_changed.connect(self.archiveTab.populate_from_profile)
        self.repoTab.repo_changed.connect(self.scheduleTab.schedulePage.populate_from_profile)
        self.repoTab.repo_added.connect(self.archiveTab.refresh_archive_list)
        self.miscTab.refresh_archive.connect(self.archiveTab.populate_from_profile)

        self.miscButton.clicked.connect(self.toggle_misc_visibility)
        self.createStartBtn.clicked.connect(self.app.create_backup_action)
        self.cancelButton.clicked.connect(self.app.backup_cancelled_event.emit)

        QShortcut(QKeySequence("Ctrl+W"), self).activated.connect(self.on_close_window)
        QShortcut(QKeySequence("Ctrl+Q"), self).activated.connect(self.on_close_window)

        self.app.backup_started_event.connect(self.backup_started_event)
        self.app.backup_finished_event.connect(self.backup_finished_event)
        self.app.backup_log_event.connect(self.set_log)
        self.app.backup_progress_event.connect(self.set_progress)
        self.app.backup_cancelled_event.connect(self.backup_cancelled_event)

        # Init profile list
        self.populate_profile_selector()
        self.profileSelector.itemClicked.connect(self.profile_clicked_action)
        self.profileSelector.currentItemChanged.connect(self.profile_selection_changed_action)
        self.profileRenameButton.clicked.connect(self.profile_rename_action)
        self.profileExportButton.clicked.connect(self.profile_export_action)
        self.profileDeleteButton.clicked.connect(self.profile_delete_action)
        self.profileAddButton.addAction(self.tr("Create new profile"), self.profile_add_action)
        self.profileAddButton.addAction(self.tr("Import from file…"), self.profile_import_action)

        # OS-specific startup options:
        if not get_network_status_monitor().is_network_status_available():
            # Hide Wifi-rule section in schedule tab.
            self.scheduleTab.networksPage.wifiListLabel.hide()
            self.scheduleTab.networksPage.wifiListWidget.hide()
            self.scheduleTab.page_2.hide()
            self.scheduleTab.toolBox.removeItem(1)

        # Connect to existing thread.
        if self.app.jobs_manager.is_worker_running():
            self.createStartBtn.setEnabled(False)
            self.createStartBtn.start()
            self.cancelButton.setEnabled(True)

        # Connect to palette change
        self._palette_connection = QApplication.instance().paletteChanged.connect(lambda p: self.set_icons())
        self.destroyed.connect(self._on_destroyed)

        self.set_icons()
        self.loaded.emit()

    def on_close_window(self):
        self.close()

    def get_current_profile(self):
        return BackupProfileModel.get(id=self.current_profile.id)

    def set_icons(self):
        self.profileAddButton.setIcon(get_colored_icon('plus'))
        self.profileRenameButton.setIcon(get_colored_icon('edit'))
        self.profileExportButton.setIcon(get_colored_icon('file-import-solid'))
        self.profileDeleteButton.setIcon(get_colored_icon('minus'))
        self.miscButton.setIcon(get_colored_icon('settings_wheel'))

    def _on_destroyed(self):
        try:
            QApplication.instance().paletteChanged.disconnect(self._palette_connection)
        except (TypeError, RuntimeError):
            pass

    def set_progress(self, text=''):
        self.progressText.setText(text)
        self.progressText.repaint()

    def set_log(self, text='', context=None):
        # Truncate very long messages to keep the status area readable
        if text and len(text) > 300:
            text = text[:300] + '...'
        self.logText.setText(text)
        self.logText.repaint()

    def _toggle_buttons(self, create_enabled=True):
        if create_enabled:
            self.createStartBtn.stop()
        else:
            self.createStartBtn.start()
        self.createStartBtn.setEnabled(create_enabled)
        self.createStartBtn.repaint()
        self.cancelButton.setEnabled(not create_enabled)
        self.cancelButton.repaint()

    def populate_profile_selector(self):
        # Clear the previous entries
        self.profileSelector.clear()

        # Keep track of the current item to be selected (if any)
        current_item = None

        # Add items to the QListWidget
        profiles = BackupProfileModel.select()
        for profile in sorted(profiles, key=lambda p: (p.name.casefold(), p.name)):
            item = QListWidgetItem(profile.name)
            item.setData(Qt.ItemDataRole.UserRole, profile.id)

            self.profileSelector.addItem(item)

            if profile.id == self.current_profile.id:
                current_item = item

        # Set the current profile as selected
        if current_item:
            self.profileSelector.setCurrentItem(current_item)

    def profile_selection_changed_action(self, index):
        profile = self.profileSelector.currentItem()
        backup_profile_id = profile.data(Qt.ItemDataRole.UserRole) if profile else None
        if not backup_profile_id:
            return

        self.current_profile = BackupProfileModel.get(id=backup_profile_id)
        SettingsModel.update({SettingsModel.str_value: self.current_profile.id}).where(
            SettingsModel.key == 'previous_profile_id'
        ).execute()

        self.app.profile_changed_event.emit()

    def profile_clicked_action(self):
        if self.miscWidget.isVisible():
            self.toggle_misc_visibility()

    def profile_rename_action(self):
        backup_profile_id = self.profileSelector.currentItem().data(Qt.ItemDataRole.UserRole)
        window = EditProfileWindow(rename_existing_id=backup_profile_id)
        self.window = window  # For tests
        window.setParent(self, QtCore.Qt.WindowType.Sheet)
        window.open()
        window.profile_changed.connect(self.profile_add_edit_result)
        window.rejected.connect(lambda: self.profileSelector.setCurrentIndex(self.profileSelector.currentIndex()))

    def profile_delete_action(self):
        if self.profileSelector.count() > 1:
            to_delete_id = self.profileSelector.currentItem().data(Qt.ItemDataRole.UserRole)
            to_delete = BackupProfileModel.get(id=to_delete_id)

            msg = self.tr("Are you sure you want to delete profile '{}'?").format(to_delete.name)
            reply = QMessageBox.question(
                self,
                self.tr("Confirm deletion"),
                msg,
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                QMessageBox.StandardButton.No,
            )

            if reply == QMessageBox.StandardButton.Yes:
                to_delete.delete_instance(recursive=True)
                self.app.scheduler.remove_job(to_delete_id)  # Remove pending jobs
                self.profileSelector.takeItem(self.profileSelector.currentRow())
                self.profile_selection_changed_action(0)

        else:
            warn = self.tr("Cannot delete the last profile.")
            point = QPoint(0, int(self.profileDeleteButton.size().height() / 2))
            QToolTip.showText(self.profileDeleteButton.mapToGlobal(point), warn)

    def profile_add_action(self):
        window = AddProfileWindow()
        self.window = window  # For tests
        window.setParent(self, QtCore.Qt.WindowType.Sheet)
        window.open()
        window.profile_changed.connect(self.profile_add_edit_result)
        window.rejected.connect(lambda: self.profileSelector.setCurrentIndex(self.profileSelector.currentIndex()))

    def profile_export_action(self):
        """
        React to pressing "Export Profile" button and save current
        profile as .json file.
        """
        window = ExportWindow(profile=self.current_profile.refresh())
        self.window = window
        window.setParent(self, QtCore.Qt.WindowType.Sheet)
        window.show()

    def profile_import_action(self):
        """
        React to "Import Profile". Ask to select a .json file and import it as
        new profile.
        """

        def profile_imported_event(profile):
            QMessageBox.information(
                None,
                self.tr('Profile import successful!'),
                self.tr('Profile {} imported.').format(profile.name),
            )
            self.populate_profile_selector()
            self.app.profile_changed_event.emit()

        filename = QFileDialog.getOpenFileName(
            self,
            self.tr("Load profile"),
            str(Path.home()),
            self.tr("JSON (*.json);;All files (*)"),
        )[0]
        if filename:
            try:
                profile_export = ProfileExport.from_json(filename)
            except ImportFailedException as exception:
                QMessageBox.critical(None, self.tr('Failed to import profile'), str(exception))
                return
            window = ImportWindow(profile_export=profile_export)
            self.window = window
            window.setParent(self, QtCore.Qt.WindowType.Sheet)
            window.profile_imported.connect(profile_imported_event)
            window.show()

    def profile_add_edit_result(self, profile_name, profile_id):
        # Profile is renamed
        if self.profileSelector.currentItem().data(Qt.ItemDataRole.UserRole) == profile_id:
            profile = self.profileSelector.takeItem(self.profileSelector.currentRow())
            profile.setText(profile_name)
        # Profile is added
        else:
            profile = QListWidgetItem(profile_name)
            profile.setData(Qt.ItemDataRole.UserRole, profile_id)
        # Insert the profile at the correct position
        # Both the casefolded and the original name are used as the key to keep the order stable
        profile_key = (profile.text().casefold(), profile.text())
        row = 0
        for i in range(self.profileSelector.count()):
            item_name = self.profileSelector.item(i).text()
            if (item_name.casefold(), item_name) < profile_key:
                row += 1
        self.profileSelector.insertItem(row, profile)
        self.profileSelector.setCurrentItem(profile)

    def toggle_misc_visibility(self):
        if self.miscWidget.isVisible():
            self.miscWidget.hide()
            self.tabWidget.setCurrentIndex(0)
            self.miscButton.setStyleSheet("font-weight: normal;")
            self.tabWidget.show()
        else:
            self.tabWidget.hide()
            self.miscWidget.setCurrentIndex(0)
            self.miscButton.setStyleSheet("font-weight: bold;")
            self.miscWidget.show()

    def backup_started_event(self):
        self._toggle_buttons(create_enabled=False)
        self.archiveTab._toggle_all_buttons(enabled=False)
        self.set_log('')

    def backup_finished_event(self):
        if not self.app.jobs_manager.is_worker_running() and (
            self.archiveTab.remaining_refresh_archives == 0 or self.archiveTab.remaining_refresh_archives == 1
        ):  # Either the refresh is done or this is the last archive to refresh.
            self._toggle_buttons(create_enabled=True)
            self.archiveTab._toggle_all_buttons(enabled=True)

    def backup_cancelled_event(self):
        self._toggle_buttons(create_enabled=True)
        self.set_log(self.tr('Task cancelled'))

    def closeEvent(self, event):
        # Save window state in SettingsModel
        SettingsModel.update({SettingsModel.str_value: str(self.width())}).where(
            SettingsModel.key == 'previous_window_width'
        ).execute()
        SettingsModel.update({SettingsModel.str_value: str(self.height())}).where(
            SettingsModel.key == 'previous_window_height'
        ).execute()

        if not is_system_tray_available():
            if SettingsModel.get(key="enable_background_question").value:
                msg = QMessageBox()
                msg.setStandardButtons(QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No)
                msg.setParent(self, QtCore.Qt.WindowType.Sheet)
                msg.setText(self.tr("Should Vorta continue to run in the background?"))
                msg.button(QMessageBox.StandardButton.Yes).clicked.connect(
                    lambda: self.miscTab.save_setting("disable_background_state", True)
                )
                msg.button(QMessageBox.StandardButton.No).clicked.connect(
                    lambda: (
                        self.miscTab.save_setting("disable_background_state", False),
                        self.app.quit(),
                    )
                )
                msg.setWindowTitle(self.tr("Quit"))
                dont_show_box = QCheckBox(self.tr("Don't show this again"))
                dont_show_box.clicked.connect(lambda x: self.miscTab.save_setting("enable_background_question", not x))
                dont_show_box.setTristate(False)
                msg.setCheckBox(dont_show_box)
                msg.exec()
            elif not SettingsModel.get(key="disable_background_state").value:
                self.app.quit()
        event.accept()



from PyQt6 import QtCore, uic
from PyQt6.QtWidgets import QApplication, QDialogButtonBox, QLabel

from vorta.borg.change_passphrase import BorgChangePassJob
from vorta.keyring.abc import VortaKeyring
from vorta.utils import get_asset
from vorta.views.partials.password_input import PasswordInput

uifile = get_asset('UI/dialogs/repo/change_passphrase.ui')
ChangeBorgPassUI, ChangeBorgPassBase = uic.loadUiType(uifile)


class ChangeBorgPassphraseWindow(ChangeBorgPassBase, ChangeBorgPassUI):
    change_borg_passphrase = QtCore.pyqtSignal(dict)

    def __init__(self, profile):
        super().__init__()
        self.setupUi(self)
        self.setWindowTitle(self.tr("Change Passphrase"))
        self.setAttribute(QtCore.Qt.WidgetAttribute.WA_DeleteOnClose)
        self.result = None
        self.profile = profile

        self.setMinimumWidth(583)

        self.passwordInput = PasswordInput()

        self.repoDataFormLayout.addRow("", self.title)  # Add title inside the form
        self.repoDataFormLayout.addRow(QLabel(self.tr("Repository:")), QLabel(str(self.profile.repo.url)))
        self.passwordInput.add_form_to_layout(self.repoDataFormLayout)

        self.saveButton = self.buttonBox.button(QDialogButtonBox.StandardButton.Ok)
        self.saveButton.setText(self.tr("Update"))

        self.errorText = QLabel("")
        self.errorText.setStyleSheet("color: red")
        self.repoDataFormLayout.addRow("", self.errorText)

        self.buttonBox.rejected.connect(self.close)
        self.buttonBox.accepted.connect(self.run)

    def run(self):
        if self.passwordInput.validate():
            newPass = self.passwordInput.passwordLineEdit.text()

            params = BorgChangePassJob.prepare(self.profile, newPass)
            if params['ok']:
                self.saveButton.setEnabled(False)
                job = BorgChangePassJob(params['cmd'], params)
                job.updated.connect(self._set_status)
                job.result.connect(self.run_result)
                QApplication.instance().jobs_manager.add_job(job)
            else:
                self._set_status(params['message'])

    def _set_status(self, text):
        self.errorText.setText(text)
        self.errorText.repaint()

    def run_result(self, result):
        self.saveButton.setEnabled(True)
        if result['returncode'] == 0:
            self.change_borg_passphrase.emit(result)
            self.accept()
        else:
            self._set_status(self.tr('Unable to change the borg passphrase.'))



from PyQt6 import uic
from PyQt6.QtWidgets import QLabel, QLineEdit, QWidget

from vorta.i18n.richtext import code, escape, format_richtext, italic, link
from vorta.utils import get_asset
from vorta.views.base_tab import BaseTab


class ShellCommandsPage(BaseTab, QWidget):
    def __init__(self, parent=None, profile_provider=None):
        super().__init__(parent=parent, profile_provider=profile_provider)
        uifile = get_asset('UI/shell_commands_page.ui')
        uic.loadUi(uifile, self)

        self.preBackupCmdLineEdit: QLineEdit = self.findChild(QLineEdit, 'preBackupCmdLineEdit')
        self.postBackupCmdLineEdit: QLineEdit = self.findChild(QLineEdit, 'postBackupCmdLineEdit')
        self.createCmdLineEdit: QLineEdit = self.findChild(QLineEdit, 'createCmdLineEdit')
        self.shellCommandsHelpLabel: QLabel = self.findChild(QLabel, 'shellCommandsHelpLabel')
        self.borgCreateHelpLabel: QLabel = self.findChild(QLabel, 'borgCreateHelpLabel')
        self._set_help_texts()

        self.bind_profile_attr(self.preBackupCmdLineEdit.textEdited, 'pre_backup_cmd')
        self.bind_profile_attr(self.postBackupCmdLineEdit.textEdited, 'post_backup_cmd')
        self.bind_repo_attr(self.createCmdLineEdit.textEdited, 'create_backup_cmd')
        self.track_profile_change(call_now=True)

    def _set_help_texts(self):
        commands_template = self.shellCommandsHelpLabel.text()
        commands_sentence = format_richtext(
            escape(
                self.tr(
                    'Run custom shell commands before and after each backup. The actual backup and post-backup '
                    'command will only run, if the pre-backup command exits without error (return code 0). '
                    'Available variables: %1'
                )
            ),
            code('$repo_url, $profile_name, $profile_slug, $returncode'),
        )
        self.shellCommandsHelpLabel.setText(format_richtext(commands_template, commands_sentence))

        borg_template = self.borgCreateHelpLabel.text()
        borg_sentence = format_richtext(
            escape(self.tr('Extra arguments for %1. Possible options are listed in %2.')),
            italic('borg create'),
            link(
                'https://borgbackup.readthedocs.io/en/stable/usage/create.html',
                self.tr('the borg documentation'),
            ),
        )
        self.borgCreateHelpLabel.setText(format_richtext(borg_template, borg_sentence))

    def populate_from_profile(self):
        profile = self.profile()
        if profile.repo:
            self.createCmdLineEdit.setText(profile.repo.create_backup_cmd)
            self.createCmdLineEdit.setEnabled(True)

            self.preBackupCmdLineEdit.setText(profile.pre_backup_cmd)
            self.preBackupCmdLineEdit.setEnabled(True)

            self.postBackupCmdLineEdit.setText(profile.post_backup_cmd)
            self.postBackupCmdLineEdit.setEnabled(True)
        else:
            self.createCmdLineEdit.setEnabled(False)
            self.preBackupCmdLineEdit.setEnabled(False)
            self.postBackupCmdLineEdit.setEnabled(False)



import logging
import os
from pathlib import Path

from PyQt6 import uic
from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QFileDialog, QMessageBox

from vorta.keyring.abc import VortaKeyring
from vorta.notifications import VortaNotifications
from vorta.profile_export import ProfileExport
from vorta.store.models import BackupProfileModel  # noqa: F401
from vorta.utils import get_asset

uifile_import = get_asset('UI/dialogs/profile/export_window.ui')
ExportWindowUI, ExportWindowBase = uic.loadUiType(uifile_import)
uifile_export = get_asset('UI/dialogs/profile/import_window.ui')
ImportWindowUI, ImportWindowBase = uic.loadUiType(uifile_export)
logger = logging.getLogger(__name__)


class ExportWindow(ExportWindowBase, ExportWindowUI):
    def __init__(self, profile):
        """
        @type profile: BackupProfileModel
        """
        super().__init__()
        self.profile = profile
        self.setupUi(self)
        self.setWindowTitle(self.tr("Export Profile"))
        self.buttonBox.accepted.connect(self.run)
        self.buttonBox.rejected.connect(self.reject)

        self.keyring = VortaKeyring.get_keyring()
        profile = self.profile
        if profile.repo is None or self.keyring.get_password('vorta-repo', profile.repo.url) is None:
            self.storePassword.setCheckState(Qt.CheckState(False))
            self.storePassword.setDisabled(True)
            self.storePassword.setToolTip(self.tr('Disclose your borg passphrase (No passphrase set)'))

    def get_file(self):
        """Get targeted save file with custom extension"""
        default_file = os.path.join(Path.home(), '{}.json'.format(self.profile.name))
        file_name = QFileDialog.getSaveFileName(self, self.tr("Save profile_export"), default_file, "JSON (*.json)")[0]
        if file_name:
            if not file_name.endswith('.json'):
                file_name += '.json'
        return file_name

    def on_error(self, error, message):
        logger.error(error)
        QMessageBox.critical(None, self.tr("Error while exporting"), message)
        self.close()

    def run(self):
        """Attempt to write profile_export export to file"""
        filename = self.get_file()
        if not filename:
            return False
        profile = self.profile
        json_string = ProfileExport.from_db(profile, self.storePassword.isChecked()).to_json()
        try:
            with open(filename, 'w') as file:
                file.write(json_string)
        except (PermissionError, OSError) as e:
            self.on_error(
                e,
                self.tr('The file {} could not be created. Please choose another location.').format(filename),
            )
            return False
        else:
            notifier = VortaNotifications.pick()
            notifier.deliver(
                self.tr('Profile export successful!'),
                self.tr('Profile export written to {}.').format(filename),
                level='info',
            )
            self.close()



from PyQt6 import uic

from vorta.utils import get_asset
from vorta.views.base_tab import BaseTab
from vorta.views.log_page import LogPage
from vorta.views.networks_page import NetworksPage
from vorta.views.schedule_page import SchedulePage
from vorta.views.shell_commands_page import ShellCommandsPage
from vorta.views.utils import get_colored_icon

uifile = get_asset('UI/schedule_tab.ui')
ScheduleUI, ScheduleBase = uic.loadUiType(uifile)


class ScheduleTab(BaseTab, ScheduleBase, ScheduleUI):
    def __init__(self, parent=None, profile_provider=None):
        super().__init__(parent=parent, profile_provider=profile_provider)
        self.setupUi(parent)
        self.toolBox.setCurrentIndex(0)
        self.init_log_page()
        self.init_shell_commands_page()
        self.init_networks_page()
        self.init_schedule_page()
        self.track_palette_change(call_now=True)
        self.track_backup_finished(self.logPage.populate_logs)

    def init_log_page(self):
        self.logPage = LogPage(self, profile_provider=self.profile)
        self.logLayout.addWidget(self.logPage)
        self.logPage.show()

    def init_shell_commands_page(self):
        self.shellCommandsPage = ShellCommandsPage(self, profile_provider=self.profile)
        self.shellCommandsLayout.addWidget(self.shellCommandsPage)
        self.shellCommandsPage.show()

    def init_networks_page(self):
        self.networksPage = NetworksPage(self, profile_provider=self.profile)
        self.networksLayout.addWidget(self.networksPage)
        self.networksPage.show()

    def init_schedule_page(self):
        self.schedulePage = SchedulePage(self, profile_provider=self.profile)
        self.scheduleLayout.addWidget(self.schedulePage)
        self.schedulePage.show()

    def set_icons(self):
        self.toolBox.setItemIcon(0, get_colored_icon('clock-o'))
        self.toolBox.setItemIcon(1, get_colored_icon('wifi'))
        self.toolBox.setItemIcon(2, get_colored_icon('tasks'))
        self.toolBox.setItemIcon(3, get_colored_icon('terminal'))



from __future__ import annotations

from typing import Any, Callable

from PyQt6.QtWidgets import QApplication

from vorta.store.models import BackupProfileModel, RepoModel

ProfileProvider = Callable[[], BackupProfileModel | None]


class BaseTab:
    """
    Shared functionality for tabs and nested pages that depend on the active profile.

    The class keeps track of Qt signal connections so subclasses only need to declare
    what they listen to instead of duplicating disconnect boilerplate.
    """

    def __init__(self, parent: Any = None, profile_provider: ProfileProvider | None = None) -> None:
        super().__init__(parent)
        self.app = QApplication.instance()
        self._profile_provider = profile_provider or self._default_profile_provider
        self._tracked_connections: list[tuple[Any, Any]] = []
        self.destroyed.connect(self._cleanup_tracked_connections)

    def _default_profile_provider(self) -> BackupProfileModel | None:
        current_profile = getattr(self.window(), 'current_profile', None)
        if current_profile is None:
            return None
        return BackupProfileModel.get(id=current_profile.id)

    def current_profile(self) -> BackupProfileModel | None:
        return self._profile_provider()

    def profile(self) -> BackupProfileModel:
        profile = self.current_profile()
        if profile is None:
            raise RuntimeError("No active backup profile is available for this view.")
        return profile

    def repo(self) -> RepoModel | None:
        return self.profile().repo

    def save_profile_attr(self, attr: str, new_value: Any) -> BackupProfileModel:
        profile = self.profile()
        setattr(profile, attr, new_value)
        profile.save()
        return profile

    def save_repo_attr(self, attr: str, new_value: Any) -> RepoModel | None:
        repo = self.repo()
        if repo is None:
            return None
        setattr(repo, attr, new_value)
        repo.save()
        return repo

    def bind_profile_attr(self, signal: Any, attr: str) -> Any:
        return self.track_signal(signal, lambda new_value: self.save_profile_attr(attr, new_value))

    def bind_repo_attr(self, signal: Any, attr: str) -> Any:
        return self.track_signal(signal, lambda new_value: self.save_repo_attr(attr, new_value))

    def track_signal(self, signal: Any, slot: Callable[..., None]) -> Any:
        connection = signal.connect(slot)
        self._tracked_connections.append((signal, connection))
        return connection

    def track_palette_change(self, callback: Callable[[], None] | None = None, call_now: bool = False) -> Any:
        callback = callback or self.set_icons
        connection = self.track_signal(self.app.paletteChanged, self._without_signal_args(callback))
        if call_now:
            callback()
        return connection

    def track_profile_change(self, callback: Callable[[], None] | None = None, call_now: bool = False) -> Any:
        callback = callback or self.populate_from_profile
        connection = self.track_signal(self.app.profile_changed_event, callback)
        if call_now:
            callback()
        return connection

    def track_backup_finished(self, callback: Callable[[], None] | None = None, call_now: bool = False) -> Any:
        callback = callback or self.populate_from_profile
        connection = self.track_signal(self.app.backup_finished_event, callback)
        if call_now:
            callback()
        return connection

    @staticmethod
    def _without_signal_args(callback: Callable[[], None]) -> Callable[..., None]:
        def wrapped(*_args, **_kwargs):
            callback()

        return wrapped

    def _cleanup_tracked_connections(self) -> None:
        while self._tracked_connections:
            signal, connection = self._tracked_connections.pop()
            try:
                signal.disconnect(connection)
            except (TypeError, RuntimeError):
                pass

    def populate_from_profile(self) -> None:
        """Template method for subclasses that render the active profile."""

    def set_icons(self) -> None:
        """Template method for subclasses that respond to palette changes."""



import os

from PyQt6 import QtCore, uic
from PyQt6.QtCore import QProcess, Qt, pyqtSlot
from PyQt6.QtWidgets import QApplication, QDialogButtonBox

from vorta.i18n import trans_late, translate
from vorta.i18n.richtext import escape, format_richtext, link
from vorta.utils import get_asset

uifile = get_asset('UI/dialogs/repo/ssh_add.ui')
SSHAddUI, SSHAddBase = uic.loadUiType(uifile)


class SSHAddWindow(SSHAddBase, SSHAddUI):
    failure = QtCore.pyqtSignal(int)

    def __init__(self):
        super().__init__()

        self.key_size_help_text = trans_late(
            'Dialog',
            '2048 or 4096 for RSA, 384 or 521 for ECDSA. Fixed for Ed25519. %1.',
        )
        self.key_size_help_link = trans_late('Dialog', 'More')

        self.setupUi(self)
        self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)

        # dialogButtonBox
        self.generateButton = self.buttonBox.button(QDialogButtonBox.StandardButton.Ok)

        self.generateButton.setText(self.tr("Generate and copy to clipboard"))

        # signals
        self.buttonBox.rejected.connect(self.reject)
        self.buttonBox.accepted.connect(self.generate_key)

        self._set_key_size_help()

        self.init_format()
        self.init_length()

    def retranslateUi(self, dialog):
        """Retranslate strings in ui."""
        super().retranslateUi(dialog)

        # setupUi calls retranslateUi
        if hasattr(self, 'generateButton'):
            self.generateButton.setText(self.tr("Generate and copy to clipboard"))
        if hasattr(self, 'keySizeHelpLabel'):
            self._set_key_size_help()

    def _set_key_size_help(self):
        template = self.keySizeHelpLabel.text()
        sentence_template = translate('Dialog', self.key_size_help_text)
        link_text = translate('Dialog', self.key_size_help_link)
        sentence = format_richtext(
            escape(sentence_template),
            link('https://stribika.github.io/2015/01/04/secure-secure-shell.html', link_text, color="#0000ff"),
        )
        self.keySizeHelpLabel.setText(format_richtext(template, sentence))

    def init_format(self):
        self.formatSelect.addItem(self.tr('ED25519 (Recommended)'), 'ed25519')
        self.formatSelect.addItem(self.tr('RSA (Legacy)'), 'rsa')
        self.formatSelect.addItem(self.tr('ECDSA'), 'ecdsa')
        self.outputFileTextBox.setText('~/.ssh/id_ed25519')
        self.formatSelect.currentIndexChanged.connect(self.format_select_change)

    def format_select_change(self, index):
        new_output = f'~/.ssh/id_{self.formatSelect.currentData()}'
        self.outputFileTextBox.setText(new_output)

    def init_length(self):
        self.lengthSelect.addItem(self.tr('High (Recommended)'), ('4096', '521'))
        self.lengthSelect.addItem(self.tr('Medium'), ('2048', '384'))

    def generate_key(self):
        format = self.formatSelect.currentData()
        length = self.lengthSelect.currentData()

        if format == 'rsa':
            length = length[0]
        else:
            length = length[1]

        output_path = os.path.expanduser(self.outputFileTextBox.text())
        if os.path.isfile(output_path):
            self.errors.setText(self.tr('Key file already exists. Not overwriting.'))
        else:
            self.sshproc = QProcess(self)
            self.sshproc.finished.connect(self.generate_key_result)
            self.sshproc.start('ssh-keygen', ['-t', format, '-b', length, '-f', output_path, '-N', ''])

    @pyqtSlot(int)
    def generate_key_result(self, exit_code):
        if exit_code == 0:
            output_path = os.path.expanduser(self.outputFileTextBox.text())
            pub_key = open(output_path + '.pub').read().strip()
            clipboard = QApplication.clipboard()
            clipboard.setText(pub_key)
            self.reject()
        else:
            self.reject()
            self.failure.emit(exit_code)

    def get_values(self):
        return {
            'ssh_key': self.sshComboBox.currentData(),
            'encryption': self.encryptionComboBox.currentData(),
            'repo_url': self.repoURL.text(),
            'password': self.passwordLineEdit.text(),
        }



import os

from PyQt6.QtCore import QDir, Qt
from PyQt6.QtGui import QFileSystemModel
from PyQt6.QtWidgets import (
    QCheckBox,
    QDialog,
    QDialogButtonBox,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QMessageBox,
    QPushButton,
    QSizePolicy,
    QTreeView,
    QVBoxLayout,
)

from vorta.views.utils import get_colored_icon


class VortaFileDialog(QDialog):
    def __init__(self, parent=None, window_title='Vorta File Dialog', title='Select files and folders:'):
        super().__init__(parent)
        if parent:
            self.setParent(parent, Qt.WindowType.Sheet)

        self.setWindowTitle(self.tr(window_title))
        self.resize(600, 400)

        layout = QVBoxLayout(self)
        path_layout = QHBoxLayout()
        bottom_layout = QHBoxLayout()

        # Show hidden files checkbox
        self.showHidden = QCheckBox(self.tr('Show hidden files'))
        self.showHidden.stateChanged.connect(self.show_hidden_files)
        # Home button
        self.btnHome = QPushButton()
        self.btnHome.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        self.btnHome.setIcon(get_colored_icon('home'))
        # Up button
        self.btnUp = QPushButton()
        self.btnUp.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        self.btnUp.setIcon(get_colored_icon('folder-on-top'))
        # Path bar
        self.path_bar = QLineEdit()
        self.path_bar.setText(QDir.homePath())
        self.path_bar.textChanged.connect(self.path_changed)

        # Path bar layout
        path_layout.addWidget(self.btnHome)
        path_layout.addWidget(self.path_bar)
        path_layout.addWidget(self.btnUp)
        layout.addLayout(path_layout)

        self.label = QLabel(self.tr(title))
        layout.addWidget(self.label)

        self.model = QFileSystemModel()
        self.model.setRootPath(QDir.rootPath())

        # Allow files and directories except '.' and '..'
        self.model.setFilter(QDir.Filter.AllEntries | QDir.Filter.NoDotAndDotDot)
        self.model.directoryLoaded.connect(lambda _: self.tree.resizeColumnToContents(0))

        self.tree = QTreeView()
        self.tree.setModel(self.model)
        self.tree.setRootIndex(self.model.index(QDir.homePath()))
        self.tree.setSelectionMode(QTreeView.SelectionMode.ExtendedSelection)
        self.tree.setHeaderHidden(True)

        # Resize the width when the directory is collapsed/expanded
        self.tree.expanded.connect(lambda _: self.tree.resizeColumnToContents(0))
        self.tree.collapsed.connect(lambda _: self.tree.resizeColumnToContents(0))

        layout.addWidget(self.tree)

        buttons = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
        buttons.button(QDialogButtonBox.StandardButton.Ok).setText("Add")
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)

        bottom_layout.addWidget(self.showHidden)
        bottom_layout.addWidget(buttons)
        layout.addLayout(bottom_layout)

        # Connections
        self.btnHome.clicked.connect(self.go_home)
        self.btnUp.clicked.connect(self.go_up)

    def selected_paths(self):
        indexes = self.tree.selectionModel().selectedIndexes()
        paths = []

        # If no tree selection, use the path bar text as the selection
        if not indexes:
            path = self.path_bar.text()
            if path and os.path.exists(path):
                if not os.access(path, os.R_OK):
                    msg = QMessageBox()
                    msg.setIcon(QMessageBox.Icon.Warning)
                    msg.setWindowTitle(self.tr("Permission Denied"))
                    msg.setText(self.tr(f"You don't have read access to {path}."))
                    msg.exec()
                    return []
                return [path]
            return []

        for index in indexes:
            if index.column() == 0:
                path = self.model.filePath(index)
                # Check for read access
                if not os.access(path, os.R_OK):
                    msg = QMessageBox()
                    msg.setIcon(QMessageBox.Icon.Warning)
                    msg.setWindowTitle(self.tr("Permission Denied"))
                    msg.setText(self.tr(f"You don't have read access to {path}."))
                    msg.exec()
                    return []
                paths.append(path)
        return list(set(paths))  # remove duplicates

    def path_changed(self):
        path = self.path_bar.text()
        if os.path.exists(path):
            self.tree.setRootIndex(self.model.index(path))
            self.tree.resizeColumnToContents(0)
            self.path_bar.setStyleSheet("")
        else:
            self.path_bar.setStyleSheet("background-color: #ffcccc")

    def go_home(self):
        self.path_bar.setText(QDir.homePath())

    def go_up(self):
        current_path = self.path_bar.text()
        parent_path = os.path.dirname(current_path.rstrip(os.sep))

        # Prevent going above the root directory
        if not parent_path or not os.path.exists(parent_path):
            return

        self.path_bar.setText(parent_path)

    def show_hidden_files(self, state):
        if state:
            self.model.setFilter(QDir.Filter.AllEntries | QDir.Filter.NoDotAndDotDot | QDir.Filter.Hidden)
        else:
            self.model.setFilter(QDir.Filter.AllEntries | QDir.Filter.NoDotAndDotDot)


class VortaFileSelector:
    def __init__(self, parent=None, window_title='Vorta File Dialog', title='Select files and folders:'):
        self.parent = parent
        self.title = title
        self.window_title = window_title

    def get_paths(self):
        dialog = VortaFileDialog(self.parent, self.window_title, self.title)
        if dialog.exec():
            return dialog.selected_paths()
        return []



from vorta.utils import borg_compat

from .borg_job import BorgJob


class BorgDiffJob(BorgJob):
    def started_event(self):
        self.app.backup_started_event.emit()
        self.app.backup_progress_event.emit(
            f"[{self.params['profile_name']}] {self.tr('Requesting differences between archives…')}"
        )

    def finished_event(self, result):
        self.app.backup_finished_event.emit(result)
        self.app.backup_progress_event.emit(
            f"[{self.params['profile_name']}] {self.tr('Obtained differences between archives.')}"
        )
        self.result.emit(result)

    @classmethod
    def prepare(cls, profile, archive_name_1, archive_name_2):
        ret = super().prepare(profile)
        if not ret['ok']:
            return ret

        ret['cmd'] = ['borg', 'diff', '--info', '--log-json']
        ret['json_lines'] = False
        if borg_compat.check('DIFF_JSON_LINES'):
            ret['cmd'].append('--json-lines')
            ret['json_lines'] = True

        if borg_compat.check('V2'):
            ret['cmd'].extend(['-r', profile.repo.url, archive_name_1, archive_name_2])
        else:
            ret['cmd'].extend([f'{profile.repo.url}::{archive_name_1}', archive_name_2])

        ret['ok'] = True
        ret['archive_name_older'] = archive_name_1
        ret['archive_name_newer'] = archive_name_2

        return ret



import os
from pathlib import PurePath

from PyQt6 import QtCore, uic
from PyQt6.QtCore import QMimeData, QUrl
from PyQt6.QtWidgets import QApplication, QLayout, QMenu, QMessageBox

from vorta.i18n import trans_late, translate
from vorta.i18n.richtext import escape, format_richtext, link
from vorta.store.models import ArchiveModel, RepoModel
from vorta.utils import borg_compat, get_asset, get_private_keys, pretty_bytes
from vorta.views.dialogs.repo.repo_add import AddRepoWindow, ExistingRepoWindow
from vorta.views.dialogs.repo.repo_change_passphrase import ChangeBorgPassphraseWindow
from vorta.views.dialogs.repo.ssh import SSHAddWindow

from .base_tab import BaseTab
from .utils import get_colored_icon

uifile = get_asset('UI/repo_tab.ui')
RepoUI, RepoBase = uic.loadUiType(uifile)


class RepoTab(BaseTab, RepoBase, RepoUI):
    repo_changed = QtCore.pyqtSignal()
    repo_added = QtCore.pyqtSignal()

    def __init__(self, parent=None, profile_provider=None):
        super().__init__(parent=parent, profile_provider=profile_provider)
        self.setupUi(parent)

        self.borgbase_sentence = trans_late('Form', 'For simple and secure backup hosting, try %1.')
        self.compression_help_text = trans_late('Form', 'Help on compression types')
        self._set_link_texts()

        # Populate dropdowns
        self.copyURLbutton.clicked.connect(self.copy_URL_action)

        # init repo add button
        self.menuAddRepo = QMenu(self.bAddRepo)

        self.menuAddRepo.addAction(self.tr("New Repository…"), self.new_repo)
        self.menuAddRepo.addAction(self.tr("Existing Repository…"), self.add_existing_repo)

        self.bAddRepo.setMenu(self.menuAddRepo)

        # init repo util button
        self.menuRepoUtil = QMenu(self.bRepoUtil)

        self.menuRepoUtil.addAction(self.tr("Unlink Repository…"), self.repo_unlink_action).setIcon(
            get_colored_icon("unlink")
        )
        self.menuRepoUtil.addAction(self.tr("Change Passphrase…"), self.repo_change_passphrase_action).setIcon(
            get_colored_icon("key")
        )

        self.bRepoUtil.setMenu(self.menuRepoUtil)

        # note: it is hard to describe these algorithms with attributes like low/medium/high
        # compression or speed on a unified scale. this is not 1-dimensional and also depends
        # on the input data. so we just tell what we know for sure.
        # "auto" is used for some slower / older algorithms to avoid wasting a lot of time
        # on incompressible data.
        self.repoCompression.addItem(self.tr('LZ4 (modern, default)'), 'lz4')
        self.repoCompression.addItem(self.tr('Zstandard Level 3 (modern)'), 'zstd,3')
        self.repoCompression.addItem(self.tr('Zstandard Level 8 (modern)'), 'zstd,8')

        # zlib and lzma come from python stdlib and are there (and in borg) since long.
        # but maybe not much reason to start with these nowadays, considering zstd supports
        # a very wide range of compression levels and has great speed. if speed is more
        # important than compression, lz4 is even a little better.
        self.repoCompression.addItem(self.tr('ZLIB Level 6 (auto, legacy)'), 'auto,zlib,6')
        self.repoCompression.addItem(self.tr('LZMA Level 6 (auto, legacy)'), 'auto,lzma,6')
        self.repoCompression.addItem(self.tr('No Compression'), 'none')
        self.repoCompression.currentIndexChanged.connect(self.compression_select_action)

        self.toggle_available_compression()
        self.repoCompression.currentIndexChanged.connect(self.compression_select_action)

        self.init_ssh()
        self.sshComboBox.currentIndexChanged.connect(self.ssh_select_action)
        self.sshKeyToClipboardButton.clicked.connect(self.ssh_copy_to_clipboard_action)
        self.bAddSSHKey.clicked.connect(self.create_ssh_key)

        self.set_icons()

        # Connect to events
        self.track_palette_change()
        self.track_profile_change(call_now=True)  # needs init of ssh and compression items
        self.track_backup_finished(self.init_repo_stats)

    def _set_link_texts(self):
        borgbase_template = self.borgbaseLinkLabel.text()
        borgbase_sentence = format_richtext(
            escape(translate('Form', self.borgbase_sentence)),
            link('https://www.borgbase.com/?utm_source=vorta&utm_medium=app', 'BorgBase'),
        )
        self.borgbaseLinkLabel.setText(format_richtext(borgbase_template, borgbase_sentence))

        compression_template = self.compressionHelpLink.text()
        compression_link = link(
            'https://borgbackup.readthedocs.io/en/stable/usage/help.html#borg-help-compression',
            translate('Form', self.compression_help_text),
        )
        self.compressionHelpLink.setText(format_richtext(compression_template, compression_link))

    def set_icons(self):
        self.bAddSSHKey.setIcon(get_colored_icon("plus"))
        self.bAddRepo.setIcon(get_colored_icon("plus"))
        self.bRepoUtil.setIcon(get_colored_icon("ellipsis-v"))
        self.sshKeyToClipboardButton.setIcon(get_colored_icon('copy'))
        self.copyURLbutton.setIcon(get_colored_icon('copy'))

    def set_repos(self):
        self.repoSelector.clear()
        self.repoSelector.addItem(self.tr('No repository selected'), None)
        for repo in RepoModel.select():
            self.repoSelector.addItem(f"{repo.name + ' - ' if repo.name else ''}{repo.url}", repo.id)
            self.repoSelector.setItemData(self.repoSelector.count() - 1, repo.url, QtCore.Qt.ItemDataRole.ToolTipRole)

    def populate_from_profile(self):
        try:
            self.repoSelector.currentIndexChanged.disconnect(self.repo_select_action)
        except TypeError:  # raised when signal is not connected
            pass

        # populate repositories
        self.set_repos()

        # load profile configuration
        profile = self.profile()
        if profile.repo:
            self.repoSelector.setCurrentIndex(self.repoSelector.findData(profile.repo.id))
        else:
            self.repoSelector.setCurrentIndex(0)

        self.repoCompression.setCurrentIndex(self.repoCompression.findData(profile.compression))
        self.sshComboBox.setCurrentIndex(self.sshComboBox.findData(profile.ssh_key))
        self.init_repo_stats()

        self.repoSelector.currentIndexChanged.connect(self.repo_select_action)

    def init_repo_stats(self):
        """Set the strings of the repo stats labels."""
        # prepare translations
        na = self.tr('N/A', "Not available.")
        no_repo_selected = self.tr("Select a repository first.")
        refresh = self.tr("Try refreshing the metadata of any archive.")

        # set labels
        repo: RepoModel = self.repo()
        if repo is not None:
            # Start with every element enabled, then disable SSH-related if relevant
            for child in self.frameRepoSettings.children():
                child.setEnabled(True)
            # local repo doesn't use ssh
            ssh_enabled = repo.is_remote_repo()
            # self.bAddSSHKey.setEnabled(ssh_enabled)
            # otherwise one cannot add a ssh key for adding a repo
            self.sshComboBox.setEnabled(ssh_enabled)
            self.sshKeyToClipboardButton.setEnabled(ssh_enabled)

            # update stats
            if repo.unique_csize is not None:
                self.sizeCompressed.setText(pretty_bytes(repo.unique_csize))
                self.sizeCompressed.setToolTip('')
            else:
                self.sizeCompressed.setText(na)
                self.sizeCompressed.setToolTip(refresh)

            if repo.unique_size is not None:
                self.sizeDeduplicated.setText(pretty_bytes(repo.unique_size))
                self.sizeDeduplicated.setToolTip('')
            else:
                self.sizeDeduplicated.setText(na)
                self.sizeDeduplicated.setToolTip(refresh)

            if repo.total_size is not None:
                self.sizeOriginal.setText(pretty_bytes(repo.total_size))
                self.sizeOriginal.setToolTip('')
            else:
                self.sizeOriginal.setText(na)
                self.sizeOriginal.setToolTip(refresh)

            self.repoEncryption.setText(str(repo.encryption))
        else:
            # Compression and SSH key are only valid entries for a repo
            # Yet Add SSH key button must be enabled for bootstrapping
            for child in self.frameRepoSettings.children():
                if not isinstance(child, QLayout):
                    child.setEnabled(False)
            self.bAddSSHKey.setEnabled(True)

            # unset stats
            self.sizeCompressed.setText(na)
            self.sizeCompressed.setToolTip(no_repo_selected)

            self.sizeDeduplicated.setText(na)
            self.sizeDeduplicated.setToolTip(no_repo_selected)

            self.sizeOriginal.setText(na)
            self.sizeOriginal.setToolTip(no_repo_selected)

            self.repoEncryption.setText(na)
            self.repoEncryption.setToolTip(no_repo_selected)

        self.repo_changed.emit()

    def init_ssh(self):
        keys = get_private_keys()
        self.sshComboBox.clear()
        self.sshComboBox.addItem(self.tr('Automatically choose SSH Key (default)'), None)
        for key in keys:
            self.sshComboBox.addItem(f'{key}', key)

    def toggle_available_compression(self):
        use_zstd = borg_compat.check('ZSTD')
        for algo in ['zstd,3', 'zstd,8']:
            ix = self.repoCompression.findData(algo)
            self.repoCompression.model().item(ix).setEnabled(use_zstd)

    def ssh_select_action(self, index):
        self.save_profile_attr('ssh_key', self.sshComboBox.itemData(index))

    def create_ssh_key(self):
        """Open a dialog to create an ssh key."""
        ssh_add_window = SSHAddWindow()
        self._window = ssh_add_window  # For tests
        ssh_add_window.setParent(self, QtCore.Qt.WindowType.Sheet)
        ssh_add_window.rejected.connect(self.init_ssh)
        ssh_add_window.failure.connect(self.create_ssh_key_failure)
        ssh_add_window.open()

    def create_ssh_key_failure(self, exit_code):
        msg = QMessageBox()
        msg.setStandardButtons(QMessageBox.StandardButton.Ok)
        msg.setParent(self, QtCore.Qt.WindowType.Sheet)
        msg.setText(self.tr(f'Error during key generation. Exited with code {exit_code}.'))
        msg.show()

    def ssh_copy_to_clipboard_action(self):
        msg = QMessageBox()
        msg.setStandardButtons(QMessageBox.StandardButton.Ok)
        msg.setParent(self, QtCore.Qt.WindowType.Sheet)

        index = self.sshComboBox.currentIndex()
        if index > 0:
            ssh_key_filename = self.sshComboBox.itemData(index)
            ssh_key_path = os.path.expanduser(f'~/.ssh/{ssh_key_filename}.pub')
            if os.path.isfile(ssh_key_path):
                pub_key = open(ssh_key_path).read().strip()
                clipboard = QApplication.clipboard()
                clipboard.setText(pub_key)

                msg.setWindowTitle(self.tr("Public Key Copied to Clipboard"))
                msg.setText(
                    self.tr(
                        "The selected public SSH key was copied to the clipboard. "
                        "Use it to set up remote repo permissions."
                    )
                )
            else:
                msg.setText(self.tr("Could not find public key."))
        else:
            msg.setText(self.tr("Select a public key from the dropdown first."))
        msg.show()

    def compression_select_action(self, index):
        self.save_profile_attr('compression', self.repoCompression.currentData())

    def new_repo(self):
        """Open a dialog to create a new repo and add it to vorta."""
        window = AddRepoWindow()
        self._window = window  # For tests
        window.setParent(self, QtCore.Qt.WindowType.Sheet)
        window.added_repo.connect(self.process_new_repo)
        # window.rejected.connect(lambda: self.repoSelector.setCurrentIndex(0))
        window.open()

    def add_existing_repo(self):
        """Open a dialog to add a existing repo to vorta."""
        window = ExistingRepoWindow()
        self._window = window  # For tests
        window.setParent(self, QtCore.Qt.WindowType.Sheet)
        window.added_repo.connect(self.process_new_repo)
        # window.rejected.connect(lambda: self.repoSelector.setCurrentIndex(0))
        window.open()

    def repo_select_action(self):
        self.save_profile_attr('repo', self.repoSelector.currentData())
        self.init_repo_stats()

    def process_new_repo(self, result):
        if result['returncode'] == 0:
            new_repo = RepoModel.get(url=result['params']['repo_url'])
            self.save_profile_attr('repo', new_repo.id)

            self.set_repos()
            self.repoSelector.setCurrentIndex(self.repoSelector.count() - 1)
            self.repo_added.emit()
            self.init_repo_stats()

    def repo_unlink_action(self):
        selected_repo_id = self.repoSelector.currentData()
        selected_repo_index = self.repoSelector.currentIndex()

        if not selected_repo_id:
            # QComboBox is empty / repo unset
            return

        # Disconnect signal before manipulating dropdown to prevent cascading updates
        try:
            self.repoSelector.currentIndexChanged.disconnect(self.repo_select_action)
        except TypeError:
            pass

        # Update database
        profile = self.profile()
        repo = RepoModel.get(id=selected_repo_id)
        repo_deleted = not repo.is_shared_with_other_profiles(excluding_profile_id=profile.id)
        profile.repo = None
        profile.save()
        if repo_deleted:
            repo.delete_instance(recursive=True)

        # Update dropdown only if the repository no longer exists in the DB.
        if repo_deleted:
            self.repoSelector.removeItem(selected_repo_index)
        self.repoSelector.setCurrentIndex(0)

        msg = QMessageBox()
        msg.setStandardButtons(QMessageBox.StandardButton.Ok)
        msg.setParent(self, QtCore.Qt.WindowType.Sheet)
        if repo_deleted:
            msg.setWindowTitle(self.tr('Repository was Unlinked'))
            msg.setText(self.tr('You can always connect it again later.'))
        else:
            msg.setWindowTitle(self.tr('Repository was Detached'))
            msg.setText(self.tr('The repository remains available for other profiles.'))
        msg.show()

        self.repo_changed.emit()
        self.populate_from_profile()

    def copy_URL_action(self):
        selected_repo_id = self.repoSelector.currentData()
        if not selected_repo_id:
            # QComboBox is empty / repo unset
            return

        repo = RepoModel.get(id=selected_repo_id)

        url = repo.url
        data = QMimeData()

        if not repo.is_remote_repo():
            path = PurePath(url)
            data.setUrls([QUrl(path.as_uri())])
            data.setText(str(path))
        else:
            data.setText(url)

        QApplication.clipboard().setMimeData(data)

    def repo_change_passphrase_action(self):
        if not self.profile().repo.encryption.startswith('repokey'):
            msg = QMessageBox()
            msg.setStandardButtons(QMessageBox.StandardButton.Ok)
            msg.setIcon(QMessageBox.Icon.Warning)
            msg.setParent(self, QtCore.Qt.WindowType.Sheet)
            msg.setWindowTitle(self.tr("Invalid Encryption Type"))
            msg.setText(self.tr("Unable to change the repository passphrase. Encryption type must be repokey."))
            msg.show()
            return

        window = ChangeBorgPassphraseWindow(self.profile())
        self._window = window
        window.setParent(self, QtCore.Qt.WindowType.Sheet)

        window.change_borg_passphrase.connect(self._handle_passphrase_change_result)
        window.open()

    def _handle_passphrase_change_result(self, result):
        """Handle the result of the passphrase change action."""
        msg = QMessageBox()
        msg.setStandardButtons(QMessageBox.StandardButton.Ok)
        msg.setParent(self, QtCore.Qt.WindowType.Sheet)

        if result['returncode'] == 0:
            msg.setIcon(QMessageBox.Icon.Information)
            msg.setWindowTitle(self.tr("Passphrase Changed"))
            msg.setText(self.tr("The borg passphrase was successfully changed."))
        else:
            msg.setIcon(QMessageBox.Icon.Warning)
            msg.setWindowTitle(self.tr("Passphrase Change Failed"))
            msg.setText(self.tr("Unable to change the repository passphrase. Please try again."))

        msg.show()



import enum
import json
import logging
import re
from dataclasses import dataclass
from pathlib import PurePath
from typing import List, Optional, Tuple

from PyQt6 import uic
from PyQt6.QtCore import (
    QDateTime,
    QLocale,
    QMimeData,
    QModelIndex,
    QPoint,
    Qt,
    QThread,
    QUrl,
)
from PyQt6.QtGui import QColor, QKeySequence, QShortcut
from PyQt6.QtWidgets import QApplication, QHeaderView, QMenu, QTreeView

from vorta.store.models import SettingsModel
from vorta.utils import get_asset, pretty_bytes, uses_dark_mode
from vorta.views.partials.treemodel import (
    FileSystemItem,
    FileTreeModel,
    FileTreeSortProxyModel,
    path_to_str,
    relative_path,
)
from vorta.views.utils import get_colored_icon

uifile = get_asset('UI/dialogs/archive/diff_result.ui')
DiffResultUI, DiffResultBase = uic.loadUiType(uifile)

logger = logging.getLogger(__name__)


class ParseThread(QThread):
    """A thread parsing diff results."""

    def __init__(self, fs_data: str, json_lines: bool, model: 'DiffTree', parent=None):
        """Init."""
        super().__init__(parent)
        self.model = model
        self.fs_data = fs_data
        self.json_lines = json_lines

    def run(self) -> None:
        """Do the work"""
        # Older version do not support json output
        if self.json_lines:
            # If fs_data is already a dict, then there was just a single json-line
            # and the default handler already parsed into json dict, otherwise
            # fs_data is a str, and needs to be split and parsed into json dicts
            if isinstance(self.fs_data, dict):
                lines = [self.fs_data]
            else:
                lines = [json.loads(line) for line in self.fs_data.split('\n') if line]

            parse_diff_json(lines, self.model)
        else:
            lines = [line for line in self.fs_data.split('\n') if line]
            parse_diff_lines(lines, self.model)


class DiffResultDialog(DiffResultBase, DiffResultUI):
    """Display the results of `borg diff`."""

    def __init__(self, archive_newer, archive_older, model: 'DiffTree'):
        """Init."""
        super().__init__()
        self.setupUi(self)

        self.model = model
        self.model.setParent(self)

        self.treeView: QTreeView
        self.treeView.setUniformRowHeights(True)  # Allows for scrolling optimizations.
        self.treeView.setAlternatingRowColors(True)
        self.treeView.setTextElideMode(Qt.TextElideMode.ElideMiddle)  # to better see name of paths

        # custom context menu
        self.treeView.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.treeView.customContextMenuRequested.connect(self.treeview_context_menu)

        # shortcuts
        shortcut_copy = QShortcut(QKeySequence.StandardKey.Copy, self.treeView)
        shortcut_copy.activated.connect(self.diff_item_copy)

        # add sort proxy model
        self.sortproxy = DiffSortProxyModel(self)
        self.sortproxy.setSourceModel(self.model)
        self.treeView.setModel(self.sortproxy)
        self.sortproxy.sorted.connect(self.slot_sorted)

        self.treeView.setSortingEnabled(True)

        # header
        header = self.treeView.header()
        header.setStretchLastSection(False)  # stretch only first section
        header.setSectionResizeMode(0, QHeaderView.ResizeMode.Stretch)
        header.setSectionResizeMode(1, QHeaderView.ResizeMode.ResizeToContents)
        header.setSectionResizeMode(2, QHeaderView.ResizeMode.ResizeToContents)

        # signals

        self.archiveNameLabel_1.setText(f'{archive_newer.name}')
        self.archiveNameLabel_2.setText(f'{archive_older.name}')

        self.comboBoxDisplayMode.currentIndexChanged.connect(self.change_display_mode)
        diff_result_display_mode = SettingsModel.get(key='diff_files_display_mode').str_value
        self.comboBoxDisplayMode.setCurrentIndex(int(diff_result_display_mode))
        self.bFoldersOnTop.toggled.connect(self.sortproxy.keepFoldersOnTop)
        self.bCollapseAll.clicked.connect(self.treeView.collapseAll)

        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

        self.set_icons()

        # Connect to palette change
        self._palette_connection = QApplication.instance().paletteChanged.connect(lambda p: self.set_icons())
        self.destroyed.connect(self._on_destroyed)

    def set_icons(self):
        """Set or update the icons in the right color scheme."""
        self.bCollapseAll.setIcon(get_colored_icon('angle-up-solid'))
        self.bFoldersOnTop.setIcon(get_colored_icon('folder-on-top'))
        self.comboBoxDisplayMode.setItemIcon(0, get_colored_icon("view-list-tree"))
        self.comboBoxDisplayMode.setItemIcon(1, get_colored_icon("view-list-tree"))
        self.comboBoxDisplayMode.setItemIcon(2, get_colored_icon("view-list-details"))

    def _on_destroyed(self):
        try:
            QApplication.instance().paletteChanged.disconnect(self._palette_connection)
        except (TypeError, RuntimeError):
            pass

    def treeview_context_menu(self, pos: QPoint):
        """Display a context menu for `treeView`."""
        index = self.treeView.indexAt(pos)
        if not index.isValid():
            # popup only for items
            return

        menu = QMenu(self.treeView)

        menu.addAction(
            get_colored_icon('copy'),
            self.tr("Copy"),
            lambda: self.diff_item_copy(index),
        )

        if self.model.getMode() != self.model.DisplayMode.FLAT:
            menu.addSeparator()
            menu.addAction(
                get_colored_icon('angle-down-solid'),
                self.tr("Expand recursively"),
                lambda: self.treeView.expandRecursively(index),
            )

        menu.popup(self.treeView.viewport().mapToGlobal(pos))

    def diff_item_copy(self, index: QModelIndex = None):
        """
        Copy a diff item path to the clipboard.

        Copies the first selected item if no index is specified.
        """
        if index is None or (not index.isValid()):
            indexes = self.treeView.selectionModel().selectedRows()

            if not indexes:
                return

            index = indexes[0]

        index = self.sortproxy.mapToSource(index)
        item: DiffItem = index.internalPointer()
        path = PurePath('/', *item.path)

        data = QMimeData()
        data.setUrls([QUrl(path.as_uri())])
        data.setText(str(path))

        QApplication.clipboard().setMimeData(data)

    def change_display_mode(self, selection: int):
        """
        Change the display mode of the tree view

        The `selection` parameter specifies the index of the selected mode in
        `comboBoxDisplayMode`.

        """
        if selection == 0:
            mode = FileTreeModel.DisplayMode.TREE
            self.bCollapseAll.setEnabled(True)
        elif selection == 1:
            mode = FileTreeModel.DisplayMode.SIMPLIFIED_TREE
            self.bCollapseAll.setEnabled(True)
        elif selection == 2:
            mode = FileTreeModel.DisplayMode.FLAT
            self.bCollapseAll.setEnabled(False)
        else:
            raise Exception("Unknown item in comboBoxDisplayMode with index {}".format(selection))

        SettingsModel.update({SettingsModel.str_value: str(selection)}).where(
            SettingsModel.key == 'diff_files_display_mode'
        ).execute()

        self.model.setMode(mode)

    def slot_sorted(self, column, order):
        """React the tree view being sorted."""
        # reveal selection
        selectedRows = self.treeView.selectionModel().selectedRows()
        if selectedRows:
            self.treeView.scrollTo(selectedRows[0])


# ---- Output parsing --------------------------------------------------------


def parse_diff_json(diffs: List[dict], model: 'DiffTree'):
    """Parse the json output from `borg diff`."""
    items = []
    for item in diffs:
        path = PurePath(item['path'])
        file_type = FileType.FILE
        changed_size = 0
        size = 0
        change_type: ChangeType = None
        mode_change: Optional[Tuple[str, str]] = None
        owner_change: Optional[Tuple[str, str, str, str]] = None
        ctime_change: Optional[Tuple[QDateTime, QDateTime]] = None
        mtime_change: Optional[Tuple[QDateTime, QDateTime]] = None
        modified: Optional[Tuple[int, int]] = None

        # added link, removed link, changed link
        # modified (added, removed), added (size), removed (size)
        # added directory, removed directory
        # owner (old_user, new_user, old_group, new_group))
        # mode (old_mode, new_mode)
        # ctime (old_ctime, new_ctime)
        # mtime (old_mtime, new_mtime)
        for change in item['changes']:
            # if more than one type of change has happened for this file/dir/link, then report the most important
            # (higher priority)
            if {'type': 'modified'} == change:
                # modified, but can't compare ids -> no added, removed
                change_type = ChangeType.MODIFIED
            elif change['type'] == 'modified':
                # modified with added, removed
                change_type = ChangeType.MODIFIED
                size = change['added'] - change['removed']
                modified = (change['added'], change['removed'])
                changed_size = sum(modified)

            elif change['type'] == 'changed link':
                change_type = ChangeType.CHANGED_LINK
                file_type = FileType.LINK

            elif change['type'].startswith(('added', 'removed')):
                if 'directory' in change['type']:
                    file_type = FileType.DIRECTORY
                elif 'link' in change['type']:
                    file_type = FileType.LINK
                elif 'chrdev' in change['type']:
                    file_type = FileType.CHRDEV
                elif 'blkdev' in change['type']:
                    file_type = FileType.BLKDEV
                elif 'fifo' in change['type']:
                    file_type = FileType.FIFO
                # else default FileType.FILE

                size = change.get('size', 0)

                a_r = change['type'].split()[0]  # 'added' or 'removed'
                if a_r == 'added':
                    change_type = ChangeType.ADDED
                else:
                    change_type = ChangeType.REMOVED
                    size = -size

                changed_size = size

            elif change['type'] == 'mode':
                # mode change can occur along with previous changes
                change_type = ChangeType.MODIFIED
                mode_change = (change['old_mode'], change['new_mode'])

            elif change['type'] == 'owner':
                # owner change can occur along with previous changes
                change_type = ChangeType.MODIFIED

                owner_change = (
                    change['old_user'],
                    change['old_group'],
                    change['new_user'],
                    change['new_group'],
                )

            elif change['type'] == 'ctime':
                # ctime change can occur along with previous changes
                change_type = ChangeType.MODIFIED
                ctime_change = (
                    QDateTime.fromString(change['old_ctime'], Qt.DateFormat.ISODateWithMs),
                    QDateTime.fromString(change['new_ctime'], Qt.DateFormat.ISODateWithMs),
                )
            elif change['type'] == 'mtime':
                # mtime change can occur along with previous changes
                change_type = ChangeType.MODIFIED
                mtime_change = (
                    QDateTime.fromString(change['old_mtime'], Qt.DateFormat.ISODateWithMs),
                    QDateTime.fromString(change['new_mtime'], Qt.DateFormat.ISODateWithMs),
                )

            else:
                raise Exception('Unknown change type: {}'.format(change['type']))

        items.append(
            (
                path,
                DiffData(
                    file_type=file_type,
                    change_type=change_type,
                    changed_size=changed_size,
                    size=size,
                    mode_change=mode_change,
                    owner_change=owner_change,
                    ctime_change=ctime_change,
                    mtime_change=mtime_change,
                    modified=modified,
                ),
            )
        )

    model.addItems(items)


# re pattern
pattern_otypes = r'directory|link|fifo|chrdev|blkdev'
pattern_ar = r'(?P<a_r>added|removed) ((?P<ar_type>' + pattern_otypes + r')|\s+(?P<size>[\d.]+) (?P<size_unit>\w+))\s*'
pattern_cl = r'changed link\s*'
pattern_modified = r'\s*\+?(?P<added>[\d.]+) (?P<added_unit>\w+)\s*-?(?P<removed>[\d.]+) (?P<removed_unit>\w+)'
pattern_mode = r'\[(?P<old_mode>[\w-]{10}) -> (?P<new_mode>[\w-]{10})\]'
pattern_owner = r'\[(?P<old_user>[\w ]+):(?P<old_group>[\w ]+) -> (?P<new_user>[\w ]+):(?P<new_group>[\w ]+)\]'
pattern_path = r'(?P<path>.*)'
pattern_changed_file = (
    r'(({ar} )|((?P<cl>{cl} )|'
    + r'((?P<modified>{modified}\s+)?)(?P<owner>{owner}\s+)?(?P<mode>{mode}\s+)?))'
    + r'{path}'
).format(
    ar=pattern_ar,
    cl=pattern_cl,
    modified=pattern_modified,
    mode=pattern_mode,
    owner=pattern_owner,
    path=pattern_path,
)
re_changed_file = re.compile(pattern_changed_file)


def parse_diff_lines(lines: List[str], model: 'DiffTree'):
    """
    Parse non-json diff output from borg.

    ::

        [-rw-rw-r-- -> lrwxrwxrwx] home/theuser/Documents/testdir/file2
        [-rw-rw-r-- -> drwxr-xr-x] home/theuser/Documents/testdir/file3
            +32 B     -36 B [-r--rw---- -> -rwxrwx--x] home/theuser/Documents/testfile.txt
        [drwxrwxr-x -> lrwxrwxrwx] home/theuser/Documents/testlink
        added directory     home/theuser/Documents/newfolder
        removed         0 B home/theuser/Documents/testdir/file1
        added          20 B home/theuser/Documents/testdir/file4
        changed link        home/theuser/Documents/testlink
        changed link [theuser:dip -> theuser:theuser] home/theuser/Documents/testlink

    Notes
    -----
    This method can't handle changes of type `modified` that do not provide
    the amount of `added` and `removed` bytes.

    """
    items = []
    for line in lines:
        if not line:
            continue

        parsed_line = re_changed_file.fullmatch(line)

        if not parsed_line:
            raise Exception("Couldn't parse diff output `{}`".format(line))

        path = PurePath(parsed_line['path'])
        file_type = FileType.FILE
        size = 0
        changed_size = 0
        change_type: ChangeType = None
        mode_change: Optional[Tuple[str, str]] = None
        owner_change: Optional[Tuple[str, str, str, str]] = None
        modified: Optional[Tuple[int, int]] = None

        if parsed_line['a_r']:
            # added or removed
            if parsed_line['ar_type']:
                if parsed_line['ar_type'] == 'directory':
                    file_type = FileType.DIRECTORY
                elif parsed_line['ar_type'] == 'link':
                    file_type = FileType.LINK
                elif parsed_line['ar_type'] == 'chrdev':
                    file_type = FileType.CHRDEV
                elif parsed_line['ar_type'] == 'blkdev':
                    file_type = FileType.BLKDEV
                elif parsed_line['ar_type'] == 'fifo':
                    file_type = FileType.FIFO
                else:
                    raise ValueError(f"Unknown file type `{parsed_line['ar_type']}`")
            else:
                # normal file
                size = size_to_byte(parsed_line['size'], parsed_line['size_unit'])

            if parsed_line['a_r'] == 'added':
                change_type = ChangeType.ADDED
            elif parsed_line['a_r'] == 'removed':
                change_type = ChangeType.REMOVED
                size = -size

            changed_size = size
        else:
            change_type = ChangeType.MODIFIED

            if parsed_line['owner']:
                # owner changed
                owner_change = (
                    parsed_line['old_user'],
                    parsed_line['old_group'],
                    parsed_line['new_user'],
                    parsed_line['new_group'],
                )

            if parsed_line['cl']:
                # link changed
                # links can't have changed permissions
                change_type = ChangeType.CHANGED_LINK
                file_type = FileType.LINK
            else:
                # modified contents or mode
                if parsed_line['modified']:
                    modified = (
                        size_to_byte(parsed_line['added'], parsed_line['added_unit']),
                        size_to_byte(parsed_line['removed'], parsed_line['removed_unit']),
                    )

                    size = modified[0] - modified[1]
                    changed_size = sum(modified)

                if parsed_line['mode']:
                    mode_change = (parsed_line['old_mode'], parsed_line['new_mode'])

        # add change to items list
        items.append(
            (
                path,
                DiffData(
                    file_type=file_type,
                    change_type=change_type,
                    changed_size=changed_size,
                    size=size,
                    mode_change=mode_change,
                    owner_change=owner_change,
                    modified=modified,
                ),
            )
        )

    model.addItems(items)


def size_to_byte(significand: str, unit: str) -> int:
    """Convert a size with a unit identifier from borg into a number of bytes."""
    if unit == 'B':
        return int(significand)
    elif unit == 'kB' or unit == 'KB':
        return int(float(significand) * 10**3)
    elif unit == 'MB':
        return int(float(significand) * 10**6)
    elif unit == 'GB':
        return int(float(significand) * 10**9)
    elif unit == 'TB':
        return int(float(significand) * 10**12)
    else:
        # unknown identifier
        raise Exception("Unknown unit `{}`".format(unit))


# ---- Sorting ---------------------------------------------------------------


class DiffSortProxyModel(FileTreeSortProxyModel):
    """
    Sort a DiffTree model.
    """

    def choose_data(self, index: QModelIndex):
        """Choose the data of index used for comparison."""
        item: DiffItem = index.internalPointer()
        if item is None or item.data is None:
            return ""
        column = index.column()

        if column == 0:
            return self.extract_path(index)
        elif column == 1:
            # change type
            ct = item.data.change_type
            if ct == ChangeType.NONE:
                return ChangeType.MODIFIED
            return ct
        elif column == 2:
            return item.data.changed_size
        else:
            # size
            return item.data.size


# ---- DiffTree --------------------------------------------------------------


class ChangeType(enum.Enum):
    """
    The possible types of changes from `borg diff`.

    modified - file contents changed.
    added - the file was added.
    removed - the file was removed.
    added directory - the directory was added.
    removed directory - the directory was removed.
    added link - the symlink was added.
    removed link - the symlink was removed.
    changed link - the symlink target was changed.
    mode - the file/directory/link mode was changed.
            Note: this could indicate a change from a file/directory/link
                    type to a different type (file/directory/link),
                    such as - a file is deleted and replaced with
                    a directory of the same name.
    owner - user and/or group ownership changed.
    ctime - creation time changed.
    mtime - modification time changed.

    size:
        If type == `added` or `removed`,
        then size provides the size of the added or removed file.
    added:
        If type == `modified` and chunk ids can be compared,
        then added and removed indicate the amount of
        data `added` and `removed`. If chunk ids can not be compared,
        then added and removed properties are not provided and
        the only information available is that the file contents were modified.
    removed:
        See added property.
    old_mode:
        If type == `mode`, then old_mode and new_mode provide the mode
        and permissions changes.
    new_mode:
        See old_mode property.
    old_user:
        If type == `owner`, then old_user, new_user, old_group
        and new_group provide the user and group ownership changes.
    old_group:
        See old_user property.
    new_user:
        See old_user property.
    new_group:
        See old_user property.
    old_ctime:
        If type == `ctime`, then old_ctime and new_ctime provide creation time changes.
    new_ctime:
        See old_ctime property.
    old_mtime:
        If type == `mtime`, then old_mtime and new_mtime provide modification time changes.
    new_mtime:
        See old_mtime property.
    """

    NONE = 0  # no change
    MODIFIED = 2  # int for sorting
    ADDED = 1
    REMOVED = 3
    ADDED_DIR = ADDED
    REMOVED_DIR = REMOVED
    ADDED_LINK = ADDED
    REMOVED_LINK = REMOVED
    CHANGED_LINK = MODIFIED
    MODE = MODIFIED  # changed permissions
    OWNER = MODIFIED
    CTIME = MODIFIED
    MTIME = MODIFIED

    def short(self):
        """Get a short identifier for the change type."""
        if self == self.ADDED:
            return 'A'
        if self == self.REMOVED:
            return 'D'
        if self == self.MODIFIED:
            return 'M'
        return ''

    def __ge__(self, other):
        """Greater than or equal for enums."""
        if self.__class__ is other.__class__:
            return other >= self
        return NotImplemented

    def __gt__(self, other):
        """Greater than for enums."""
        if self.__class__ is other.__class__:
            return other < self
        return NotImplemented

    def __le__(self, other):
        """Lower than or equal for enums."""
        if self.__class__ is other.__class__:
            return self.value == other.value or self < other
        return NotImplemented

    def __lt__(self, other):
        """Lower than for enums."""
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented


class FileType(enum.Enum):
    """The possible file types of changed file."""

    FILE = enum.auto()
    DIRECTORY = enum.auto()
    LINK = enum.auto()
    CHRDEV = enum.auto()
    BLKDEV = enum.auto()
    FIFO = enum.auto()


@dataclass
class DiffData:
    """The data linked to a diff item."""

    file_type: FileType
    change_type: ChangeType
    changed_size: int  # total modified bits
    size: int  # size change (disk usage)
    mode_change: Optional[Tuple[str, str]] = None
    owner_change: Optional[Tuple[str, str, str, str]] = None
    ctime_change: Optional[Tuple[QDateTime, QDateTime]] = None
    mtime_change: Optional[Tuple[QDateTime, QDateTime]] = None
    modified: Optional[Tuple[int, int]] = None


DiffItem = FileSystemItem[DiffData]


class DiffTree(FileTreeModel[DiffData]):
    """The file tree model for diff results."""

    def _make_filesystemitem(self, path, data):
        return super()._make_filesystemitem(path, data)

    def _merge_data(self, item, data):
        if data:
            logger.debug('Overriding data for {}'.format(path_to_str(item.path)))
        return super()._merge_data(item, data)

    def _flat_filter(self, item):
        """
        Return whether an item is part of the flat model representation.

        The item's data might have not been set yet.
        """
        return item.data and item.data.change_type != ChangeType.NONE

    def _simplify_filter(self, item: DiffItem) -> bool:
        """
        Return whether an item may be merged in simplified mode.

        Allows simplification only for unchanged items.
        """
        for child in item.children:
            if child.data.change_type != ChangeType.NONE:
                return False

        if item.data.change_type == ChangeType.NONE:
            return True

        return False  # otherwise the change is not displayed

    def _process_child(self, child):
        """
        Process a new child.

        This can make some changes to the child's data like
        setting a default value if the child's data is None.
        This can also update the data of the parent.
        This must emit `dataChanged` if data is changed.

        Parameters
        ----------
        child : FileSystemItem
            The child that was added.
        """
        parent = child._parent

        if not child.data:
            child.data = DiffData(FileType.DIRECTORY, ChangeType.NONE, 0, 0)

        if child.data.size != 0 or child.data.changed_size != 0:
            # update size
            size = child.data.size
            changed_size = child.data.changed_size

            def add_size(parent):
                if parent is self.root:
                    return

                if parent.data is None:
                    raise Exception("Item {} without data".format(path_to_str(parent.path)))
                else:
                    parent.data.size += size
                    parent.data.changed_size += changed_size

                # update parent
                parent = parent._parent
                if parent:
                    add_size(parent)

            add_size(parent)

    def columnCount(self, parent: QModelIndex = QModelIndex()) -> int:
        """
        Returns the number of columns for the children of the given parent.

        This corresponds to the number of data (column) entries shown
        for each item in the tree view.

        Parameters
        ----------
        parent : QModelIndex, optional
            The index of the parent, by default QModelIndex()

        Returns
        -------
        int
            The number of rows.
        """
        # name, change_type, changed bytes, size balance
        return 4

    def headerData(
        self,
        section: int,
        orientation: Qt.Orientation,
        role: int = Qt.ItemDataRole.DisplayRole,
    ):
        """
        Get the data for the given role and section in the given header.

        The header is identified by its orientation.
        For horizontal headers, the section number corresponds to
        the column number. Similarly, for vertical headers,
        the section number corresponds to the row number.

        Parameters
        ----------
        section : int
            The row or column number.
        orientation : Qt.Orientation
            The orientation of the header.
        role : int, optional
            The data role, by default Qt.ItemDataRole.DisplayRole

        Returns
        -------Improve
        Any
            The data for the specified header section.
        """
        if orientation == Qt.Orientation.Horizontal and role == Qt.ItemDataRole.DisplayRole:
            if section == 0:
                return self.tr("Name")
            elif section == 1:
                return self.tr("Change")
            elif section == 2:
                return self.tr("Size")
            elif section == 3:
                return self.tr("Balance")

        return None

    def data(self, index: QModelIndex, role: int = Qt.ItemDataRole.DisplayRole):
        """
        Get the data for the given role and index.

        The indexes internal pointer references the corresponding
        `FileSystemItem`.

        Parameters
        ----------
        index : QModelIndex
            The index of the item.
        role : int, optional
            The data role, by default Qt.ItemDataRole.DisplayRole

        Returns
        -------
        Any
            The data, return None if no data is available for the role.
        """
        if not index.isValid():
            return None

        item: DiffItem = index.internalPointer()
        column = index.column()

        if role == Qt.ItemDataRole.DisplayRole:
            if column == 0:
                # name
                if self.mode == self.DisplayMode.FLAT:
                    return path_to_str(item.path)

                if self.mode == self.DisplayMode.SIMPLIFIED_TREE:
                    parent = index.parent()
                    if parent == QModelIndex():
                        return path_to_str(relative_path(self.root.path, item.path))

                    return path_to_str(relative_path(parent.internalPointer().path, item.path))

                # standard tree mode
                return item.subpath
            elif column == 1:
                # change type
                return item.data.change_type.short()
            elif column == 2:
                return pretty_bytes(item.data.changed_size)
            else:
                # size
                return pretty_bytes(item.data.size)

        if role == Qt.ItemDataRole.ForegroundRole:
            # colour
            if item.data.change_type == ChangeType.ADDED:
                return QColor(Qt.GlobalColor.green) if uses_dark_mode() else QColor(Qt.GlobalColor.darkGreen)
            if item.data.change_type == ChangeType.MODIFIED:
                return QColor(Qt.GlobalColor.yellow) if uses_dark_mode() else QColor(Qt.GlobalColor.darkYellow)
            if item.data.change_type == ChangeType.REMOVED:
                return QColor(Qt.GlobalColor.red) if uses_dark_mode() else QColor(Qt.GlobalColor.darkRed)
            return None  # no change

        if role == Qt.ItemDataRole.ToolTipRole:
            if column == 0:
                # name column -> display fullpath
                return path_to_str(item.path)

            # info/data tooltip -> no real size limitation
            tooltip_template = "{name}\n" + "\n" + "{filetype} {changetype}"

            modified_template = self.tr("Added {}, deleted {}")
            owner_template = "{: <10} -> {: >10}"
            time_template = "{}: {} -> {}"
            permission_template = "{} -> {}"

            # format
            if item.data.file_type == FileType.FILE:
                filetype = self.tr("File")
            elif item.data.file_type == FileType.DIRECTORY:
                filetype = self.tr("Directory")
            elif item.data.file_type == FileType.LINK:
                filetype = self.tr("Link")
            elif item.data.file_type == FileType.BLKDEV:
                filetype = self.tr("Block device file")
            elif item.data.file_type == FileType.CHRDEV:
                filetype = self.tr("Character device file")
            else:
                raise Exception("Unknown filetype {}".format(item.data.file_type))

            if item.data.change_type == ChangeType.NONE:
                changetype = self.tr("unchanged")
            elif item.data.change_type == ChangeType.MODIFIED:
                changetype = self.tr("modified")
            elif item.data.change_type == ChangeType.REMOVED:
                changetype = self.tr("removed")
            elif item.data.change_type == ChangeType.ADDED:
                changetype = self.tr("added")
            else:
                raise Exception("Unknown changetype {}".format(item.data.change_type))

            tooltip = tooltip_template.format(name=item.path[-1], filetype=filetype, changetype=changetype)
            if item.data.modified:
                tooltip += '\n'
                tooltip += modified_template.format(
                    pretty_bytes(item.data.modified[0]),
                    pretty_bytes(item.data.modified[1]),
                )

            if item.data.mode_change:
                tooltip += '\n'
                tooltip += permission_template.format(*item.data.mode_change)

            if item.data.owner_change:
                tooltip += '\n'
                tooltip += owner_template.format(
                    '{}:{}'.format(item.data.owner_change[0], item.data.owner_change[1]),
                    "{}:{}".format(item.data.owner_change[2], item.data.owner_change[3]),
                )

            if item.data.ctime_change:
                tooltip += '\n'
                tooltip += time_template.format(
                    "Creation Time",
                    QLocale.system().toString(item.data.ctime_change[0], QLocale.FormatType.ShortFormat),
                    QLocale.system().toString(item.data.ctime_change[1], QLocale.FormatType.ShortFormat),
                )

            if item.data.mtime_change:
                tooltip += '\n'
                tooltip += time_template.format(
                    "Modification Time",
                    QLocale.system().toString(item.data.mtime_change[0], QLocale.FormatType.ShortFormat),
                    QLocale.system().toString(item.data.mtime_change[1], QLocale.FormatType.ShortFormat),
                )

            return tooltip
